from functools import wraps

import cfdm

from .constants import ValidLogLevels
from .functions import (
    _DEPRECATION_ERROR_KWARGS,
    _disable_logging,
    _is_valid_log_level_int,
    _reset_log_emergence_level,
    log_level,
)

# Decorators (and helper functions for these) inherited from cfdm:
_inplace_enabled = cfdm._inplace_enabled
_inplace_enabled_define_and_cleanup = cfdm._inplace_enabled_define_and_cleanup
_manage_log_level_via_verbosity = cfdm._manage_log_level_via_verbosity
_display_or_return = cfdm._display_or_return


# @_deprecated_kwarg_check('i') -> example usage for decorating, using i kwarg
def _deprecated_kwarg_check(*depr_kwargs, version=None, removed_at=None):
    """A wrapper for provision of positional arguments to the
    decorator."""

    def deprecated_kwarg_check_decorator(operation_method):
        """A decorator for a deprecation check on given kwargs.

        To specify deprecated kwargs, supply them as string arguments,
        with the version at which they were deprecated, and optionally
        the version at which the kwargs will be completely
        removed. E.g:

            @_deprecated_kwarg_check('i', version="3.0.0")
            @_deprecated_kwarg_check('i', 'traceback', version="3.0.0")
            @_deprecated_kwarg_check('i', version="3.0.0")
            @_deprecated_kwarg_check('i', version="3.0.0", removed_at="4.0.0")

        For a specified list `depr_kwargs`, check if the
        decorated method has been supplied with any of the elements as
        keyword arguments and if so, call _DEPRECATION_ERROR_KWARGS on
        them, optionally providing a custom message to raise inside
        it.

        """

        @wraps(operation_method)
        def precede_with_kwarg_deprecation_check(self, *args, **kwargs):

            # If there is only one input deprecated kwarg, form a list so the
            # following loop does not iterate over characters. This means we
            # do not have to write 'dkwarg,' instead of 'dkwarg' each time.
            if isinstance(depr_kwargs, str):
                # Note: can't simply reassign to depr_kwarg and use that in the
                # following loop to iterate over, as this leads to an
                # UnboundLocalError. Easiest way is to use another variable.
                depr_kwargs_list = [depr_kwargs]
            else:
                depr_kwargs_list = depr_kwargs

            for depr_kwarg in depr_kwargs_list:
                if kwargs.get(depr_kwarg):
                    pass_in_kwarg = {depr_kwarg: True}
                    _DEPRECATION_ERROR_KWARGS(
                        self,
                        operation_method.__name__,
                        pass_in_kwarg,
                        version=version,
                        removed_at=removed_at,
                    )  # pragma: no cover

            operation_method_result = operation_method(self, *args, **kwargs)

            # Decorated method has same return signature as if undecorated:
            return operation_method_result

        return precede_with_kwarg_deprecation_check

    return deprecated_kwarg_check_decorator


# TODO(?): we could instead define a metaclass to apply this logic to all
# methods in a class with a self.verbose attibute to avoid decorating perhaps
# most methods of the class (any that contain log calls should have it).
# But this (explicit) approach may be better?
def _manage_log_level_via_verbose_attr(method_using_verbose_attr, calls=[0]):
    """A decorator for managing log message filtering by verbose
    attribute.

    Note this has identical purpose to _manage_log_level_via_verbosity except
    it adapts the log level based on a verbose attribute of the class
    instance, rather than a verbose keyword argument to the function.

    This enables overriding of the log severity level such that an integer
    value (lying in the valid range) of self.verbose on the decorated
    method will ignore the global cf.log_level() to configure a custom
    verbosity for the individual method call, applying to its logic and any
    functions it calls internally and lasting only the duration of the call.

    If self.verbose=None, as is the default, the log_level() determines
    which log messages are shown, as standard.

    Only use this to decorate methods which make log calls directly and are
    in a class with a 'verbose' instance attribute set to None by default.

    Note that the 'calls' keyword argument is to automatically track the number
    of decorated functions that are being (or about to be) executed, with the
    purpose of preventing resetting of the effective log level at the
    completion of decorated functions that are called inside other decorated
    functions (see comments in 'finally' statement for further explanation).
    Note (when it is of concern) that this approach may not be thread-safe.

    """

    @wraps(method_using_verbose_attr)
    def verbose_override_wrapper(self, *args, **kwargs):
        # Increment indicates that one decorated function has started execution
        calls[0] += 1

        lvls = ", ".join(
            [val.name + " = " + str(val.value) for val in ValidLogLevels]
        )
        invalid_attr_msg = (
            f"Invalid value '{self.verbose}' for the 'self.verbose' "
            "attribute. Accepted values are integers corresponding in "
            f"positive cases to increasing verbosity (namely {lvls}), or "
            "None, to configure the verbosity according to the global "
            "log_level setting."
        )
        # Convert Boolean cases for backwards compatibility. Need 'is' identity
        # rather than '==' (value) equivalency test, since 1 == True, etc.
        # Note: it is safe (and desirable) to change these (to standard form)
        if self.verbose is True:
            self.verbose = 3  # max verbosity excluding debug levels
        elif self.verbose is False:
            self.verbose = 0  # corresponds to disabling logs i.e. no verbosity

        # May change during decorator, whereas we don't want to amend the attr:
        verbose_attr = self.verbose
        # First convert valid string inputs to the enum-mapped int constant:
        if isinstance(verbose_attr, str):
            uppercase_attr = verbose_attr.upper()
            if hasattr(ValidLogLevels, uppercase_attr):
                verbose_attr = getattr(ValidLogLevels, uppercase_attr).value
            else:
                raise ValueError(invalid_attr_msg)

        # Override log levels for the function & all it calls (to reset at end)
        if verbose_attr is not None:  # None as default; exclude True & False
            if _is_valid_log_level_int(verbose_attr):
                _reset_log_emergence_level(verbose_attr)
            else:
                raise ValueError(invalid_attr_msg)

        # First need to (temporarily) re-enable global logging if disabled
        # in the cases where you do not want to disable it anyway:
        if log_level() == "DISABLE" and verbose_attr not in (0, None):
            _disable_logging(at_level="NOTSET")  # enables all logging again

        # After method completes, re-set any changes to log level or enabling
        try:
            return method_using_verbose_attr(self, *args, **kwargs)
        except Exception:
            raise
        finally:  # so that crucial 'teardown' code runs even if method errors
            # Decrement indicates one decorated function has finished execution
            calls[0] -= 1
            # Due to the incrementing and decrementing of 'calls', it will
            # only be zero when the outermost decorated method has finished,
            # so the following condition prevents resetting occurring once
            # inner functions complete (which would mean any subsequent code
            # in the outer function would undesirably regain the global level):
            if calls[0] == 0:
                if verbose_attr == 0:
                    _disable_logging(at_level="NOTSET")  # lift deactivation
                elif verbose_attr is not None and _is_valid_log_level_int(
                    verbose_attr
                ):
                    _reset_log_emergence_level(log_level())
                if log_level() == "DISABLE" and verbose_attr != 0:
                    _disable_logging()  # disable again after re-enabling

    return verbose_override_wrapper



import cfdm

from . import mixin


class InterpolationParameter(
    mixin.PropertiesData, cfdm.InterpolationParameter
):
    """An interpolation parameter variable.

    Space may be saved by storing a subsample of the coordinates. The
    uncompressed coordinates can be reconstituted by interpolation
    from the subsampled coordinate values, also called either "tie
    points" or "bounds tie points".

    An interpolation parameter variable provides values for
    coefficient terms in the interpolation equation, or for any other
    terms that configure the interpolation process.

    **NetCDF interface**

    {{netCDF variable}}

    The netCDF subsampled dimension name and the netCDF interpolation
    subarea dimension name, if required, are set on the
    corresponding tie point index variable.

    .. versionadded:: TODODASKVER

    .. seealso:: `TiePointIndex`

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)



import cfdm

from . import mixin
from .data import Data
from .decorators import _deprecated_kwarg_check
from .units import Units


class Bounds(mixin.Coordinate, mixin.PropertiesData, cfdm.Bounds):
    """A cell bounds component of a coordinate or domain ancillary
    construct of the CF data model.

    An array of cell bounds spans the same domain axes as its
    coordinate array, with the addition of an extra dimension whose
    size is that of the number of vertices of each cell. This extra
    dimension does not correspond to a domain axis construct since it
    does not relate to an independent axis of the domain. Note that,
    for climatological time axes, the bounds are interpreted in a
    special way indicated by the cell method constructs.

    In the CF data model, a bounds component does not have its own
    properties because they can not logically be different to those of
    the coordinate construct itself. However, it is sometimes desired
    to store attributes on a CF-netCDF bounds variable, so it is also
    allowed to provide properties to a bounds component.

    **NetCDF interface**

    {{netCDF variable}}

    The name of the trailing netCDF dimension spanned by bounds (which
    does not correspond to a domain axis construct) may be accessed
    with the `nc_set_dimension`, `nc_get_dimension`,
    `nc_del_dimension` and `nc_has_dimension` methods.

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)

    def contiguous(self, overlap=True, direction=None, period=None, verbose=1):
        """Return True if the bounds are contiguous.

         Bounds are contiguous if the cell boundaries match up, or
         overlap, with the boundaries of adjacent cells.

         In general, it is only possible for 1 or 0 variable
         dimensional variables with bounds to be contiguous, but size
         1 variables with any number of dimensions are always
         contiguous.

         An exception is raised if the variable is multidimensional
         and has more than one element.

        .. versionadded:: 2.0

        :Parameters:

             overlap: `bool`, optional
                 If False then 1-d cells with two bounds vertices are
                 not considered contiguous if any adjacent cells
                 overlap each other. By default such cells are
                 considered contiguous.

             direction:
                 Specify the direction of 1-d coordinates with two
                 bounds vertices. Either True for increasing
                 coordinates, or False for decreasing coordinates. By
                 default the direction is inferred from whether the
                 first bound of the first cell is less than its second
                 bound (direction is True), or not (direction is
                 False).

             period: optional
                 Define the period of cyclic values so that the test
                 for contiguousness can be carried out with modulo
                 arithmetic. By default the data are assumed to be
                 non-cyclic, unless the bounds have units of longitude
                 (or have units of ``'degrees'``), in which case a
                 period of 360 degrees is assumed.

             {{verbose: `int` or `str` or `None`, optional}}

         :Returns:

             `bool`
                 Whether or not the cells are contiguous.

         **Examples**

         >>> c.has_bounds()
         False
         >>> c.contiguous()
         False

         >>> print(c.bounds[:, 0])
         [  0.5   1.5   2.5   3.5 ]
         >>> print(c.bounds[:, 1])
         [  1.5   2.5   3.5   4.5 ]
         >>> c.contiuous()
         True

         >>> print(c.bounds[:, 0])
         [  0.5   1.5   2.5   3.5 ]
         >>> print(c.bounds[:, 1])
         [  2.5   3.5   4.5   5.5 ]
         >>> c.contiuous()
         True
         >>> c.contiuous(overlap=False)
         False

        """
        data = self.get_data(None, _fill_value=False)
        if data is None:
            return False

        ndim = data.ndim - 1
        nbounds = data.shape[-1]

        if data.size == nbounds:
            return True

        if period is None:
            if self.Units.islongitude:
                period = Data(360.0, "degrees_east")
            elif self.Units.equals(Units("degrees")):
                period = Data(360.0, "degrees")

        if verbose >= 2:
            print(f"Period = {period!r}")

        if ndim == 2:
            if nbounds != 4:
                raise ValueError(
                    f"Can't tell if {ndim}-d cells with {nbounds} vertices "
                    "are contiguous"
                )

            # --------------------------------------------------------
            # 2-d coordinates with 4 vertices per cell
            # --------------------------------------------------------
            #            if overlap:
            #                raise ValueError(
            #                    "overlap=True and can't tell if 2-d bounds are contiguous")

            # Check cells (j, i) and cells (j, i+1) are contiguous
            diff = data[:, :-1, 1] - data[:, 1:, 0]
            if period is not None:
                diff = diff % period

            if diff.any():
                return False

            diff = data[:, :-1, 2] - data[:, 1:, 3]
            if period is not None:
                diff = diff % period

            if diff.any():
                return False

            # Check cells (j, i) and (j+1, i) are contiguous
            diff = data[:-1, :, 3] - data[1:, :, 0]
            if period is not None:
                diff = diff % period

            if diff.any():
                return False

            diff = data[:-1, :, 2] - data[1:, :, 1]
            if period is not None:
                diff = diff % period

            if diff.any():
                return False

            #            bnd = data.array
            #            for j in range(data.shape[0] - 1):
            #                for i in range(data.shape[1] - 1):
            #
            #                    if (bnd[j, i, 1] != bnd[j, i+1, 0] or
            #                        bnd[j, i, 2] != bnd[j, i+1, 3]):
            #                        return False
            #
            #                    # Check cells (j, i) and (j+1, i) are contiguous
            #                    if (bnd[j, i, 3] != bnd[j+1, i, 0] or
            #                            bnd[j, i, 2] != bnd[j+1, i, 1]):
            #                        return False

            return True

        if ndim > 2:
            raise ValueError(f"Can't tell if {ndim}-d cells are contiguous")

        if nbounds != 2:
            raise ValueError(
                f"Can't tell if {ndim}-d cells with {nbounds} vertices "
                "are contiguous"
            )

        if not overlap:
            diff = data[1:, 0] - data[:-1, 1]
            if period is not None:
                diff = diff % period

            return not diff.any()
        #            return data[1:, 0].equals(data[:-1, 1])
        else:
            if direction is None:
                b = data[(0,) * ndim].array
                direction = b.item(0) < b.item(1)

            if direction:
                return (data[1:, 0] <= data[:-1, 1]).all()
            else:
                return (data[1:, 0] >= data[:-1, 1]).all()

    @_deprecated_kwarg_check(
        "relaxed_identity", version="3.0.0", removed_at="4.0.0"
    )
    def identity(
        self,
        default="",
        strict=False,
        relaxed=False,
        nc_only=False,
        relaxed_identity=None,
    ):
        """Return the canonical identity.

        By default the identity is the first found of the following:

        1. The "standard_name" property.
        2. The "cf_role" property, preceded by ``'cf_role='``.
        3. The "long_name" property, preceded by ``'long_name='``.
        4. The netCDF variable name, preceded by ``'ncvar%'``.
        5. The value of the *default* parameter.

        Properties include any inherited properties.

        .. versionadded:: 3.0.6

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

            strict: `bool`, optional
                If True then the identity is the first found of only the
                "standard_name" property or the "id" attribute.

            relaxed: `bool`, optional
                If True then the identity is the first found of only the
                "standard_name" property, the "id" attribute, the
                "long_name" property or the netCDF variable name.

            nc_only: `bool`, optional
                If True then only take the identity from the netCDF
                variable name.

        :Returns:

                The identity.

        **Examples**

        >>> b.inherited_properties()
        {'foo': 'bar',
         'long_name': 'Longitude'}
        >>> b.properties()
        {'long_name': 'A different long name'}
        >>> b.identity()
        'long_name=A different long name'
        >>> b.del_property('long_name')
        'A different long name'
        >>> b.identity()
        'long_name=Longitude'

        """
        inherited_properties = self.inherited_properties()
        if inherited_properties:
            bounds = self.copy()
            properties = bounds.properties()
            bounds.set_properties(inherited_properties)
            bounds.set_properties(properties)
            self = bounds

        return super().identity(
            default=default, strict=strict, relaxed=relaxed, nc_only=nc_only
        )



from ..array.netcdfarray import NetCDFArray
from .abstract import FragmentArray


class NetCDFFragmentArray(FragmentArray):
    """A CFA fragment array stored in a netCDF file.

    .. versionadded:: TODODASKVER

    """

    def __init__(
        self,
        filename=None,
        address=None,
        dtype=None,
        shape=None,
        aggregated_units=False,
        aggregated_calendar=False,
        units=False,
        calendar=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            filename: `str`
                The name of the netCDF fragment file containing the
                array.

            address: `str`, optional
                The name of the netCDF variable containing the
                fragment array. Required unless *varid* is set.

            dtype: `numpy.dtype`
                The data type of the aggregated array. May be `None`
                if the numpy data-type is not known (which can be the
                case for netCDF string types, for example). This may
                differ from the data type of the netCDF fragment
                variable.

            shape: `tuple`
                The shape of the fragment within the aggregated
                array. This may differ from the shape of the netCDF
                fragment variable in that the latter may have fewer
                size 1 dimensions.

            units: `str` or `None`, optional
                The units of the fragment data. Set to `None` to
                indicate that there are no units. If unset then the
                units will be set during the first `__getitem__` call.

            calendar: `str` or `None`, optional
                The calendar of the fragment data. Set to `None` to
                indicate the CF default calendar, if applicable. If
                unset then the calendar will be set during the first
                `__getitem__` call.

            {{aggregated_units: `str` or `None`, optional}}

            {{aggregated_calendar: `str` or `None`, optional}}

            source: optional
                Initialise the array from the given object.

                {{init source}}

            {{deep copy}}

        """
        if source is not None:
            super().__init__(source=source, copy=copy)
            return

        if isinstance(address, int):
            ncvar = None
            varid = address
        else:
            ncvar = address
            varid = None

        # TODO set groups from ncvar
        group = None

        array = NetCDFArray(
            filename=filename,
            ncvar=ncvar,
            varid=varid,
            group=group,
            dtype=dtype,
            shape=shape,
            mask=True,
            units=units,
            calendar=calendar,
            copy=False,
        )

        super().__init__(
            filename=filename,
            address=address,
            dtype=dtype,
            shape=shape,
            aggregated_units=aggregated_units,
            aggregated_calendar=aggregated_calendar,
            array=array,
            source=source,
            copy=False,
        )

    def __getitem__(self, indices):
        """Returns a subspace of the fragment as a numpy array.

        x.__getitem__(indices) <==> x[indices]

        Indexing is similar to numpy indexing, with the following
        differences:

          * A dimension's index can't be rank-reducing, i.e. it can't
            be an integer, nor a scalar `numpy` or `dask` array.

          * When two or more dimension's indices are sequences of
            integers then these indices work independently along each
            dimension (similar to the way vector subscripts work in
            Fortran).

        **Performance**

        If the netCDF fragment variable has fewer than `ndim`
        dimensions then the entire array is read into memory before
        the requested subspace of it is returned.

        .. versionadded:: TODODASKVER

        """
        indices = self._parse_indices(indices)
        array = self.get_array()

        try:
            array = array[indices]
        except ValueError:
            # A value error is raised if indices has at least ndim
            # elements but the netCDF fragment variable has fewer than
            # ndim dimensions. In this case we get the entire fragment
            # array, insert the missing size 1 dimensions, and then
            # apply the requested slice.
            array = array[Ellipsis]
            if array.ndim < self.ndim:
                array = array.reshape(self.shape)

            array = array[indices]

        array = self._conform_units(array)
        return array



import fnmatch
import os
import re
import subprocess
from distutils.command.build import build

from setuptools import setup


def find_package_data_files(directory):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, "*"):
                filename = os.path.join(root, basename)
                yield filename.replace("cf/", "", 1)


def find_test_files():
    """Yield the data files in cf/test/"""
    for filename in os.listdir("cf/test"):
        if (
            fnmatch.fnmatch(filename, "*.sh")
            or fnmatch.fnmatch(filename, "*.nc")
            or fnmatch.fnmatch(filename, "*.pp")
            or fnmatch.fnmatch(filename, "*.cdl")
        ):
            filename = os.path.join("test", filename)
            yield filename


def _read(fname):
    """Returns content of a file."""
    fpath = os.path.dirname(__file__)
    fpath = os.path.join(fpath, fname)
    with open(fpath, "r") as file_:
        return file_.read()


def _get_version():
    """Returns library version by inspecting __init__.py file."""
    return re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        _read("cf/__init__.py"),
        re.MULTILINE,
    ).group(1)


version = _get_version()
packages = ["cf"]
etc_files = [f for f in find_package_data_files("cf/etc")]
umread_files = [f for f in find_package_data_files("cf/umread_lib/c-lib")]
test_files = [f for f in find_test_files()]

package_data = etc_files + umread_files + test_files


class build_umread(build):
    """Adpated from https://github.com/Turbo87/py-
    xcsoar/blob/master/setup.py."""

    def run(self):
        # Run original build code
        build.run(self)

        # Build umread
        print("Running build_umread")

        build_dir = os.path.join(
            os.path.abspath(self.build_lib), "cf/umread_lib/c-lib"
        )

        cmd = ["make", "-C", build_dir]

        #        rc = subprocess.call(cmd)

        def compile():
            print("*" * 80)
            print("Running:", " ".join(cmd), "\n")

            try:
                rc = subprocess.call(cmd)
            except Exception as error:
                print(error)
                rc = 40

            print("\n", "-" * 80)
            if not rc:
                print("SUCCESSFULLY built UM read C library")
            else:
                print("WARNING: Failed to build the UM read C library.")
                print(
                    "         Attempting to read UKMO PP and UM format files "
                    "will result in failure."
                )
                print(
                    "         This will not affect any other cf functionality."
                )
                print(
                    "         In particular, netCDF file processing is "
                    "unaffected."
                )

            print("-" * 80)
            print("\n", "*" * 80)
            print()
            print("cf build successful")
            print()

        self.execute(compile, [], "compiling umread")

    # --- End: class


long_description = """
CF Python
=========

The Python cf package is an Earth science data analysis library that
is built on a complete implementation of the CF data model


Documentation
=============

http://ncas-cms.github.io/cf-python

Tutorial
========

https://ncas-cms.github.io/cf-python/tutorial

Installation
============

http://ncas-cms.github.io/cf-python/installation


Functionality
=============

The `cf` package implements the CF data model
(https://doi.org/10.5194/gmd-10-4619-2017) for its internal data
structures and so is able to process any CF-compliant dataset. It is
not strict about CF-compliance, however, so that partially conformant
datasets may be ingested from existing datasets and written to new
datasets. This is so that datasets which are partially conformant may
nonetheless be modified in memory.

The `cf` package can:

* read field constructs and domain constructs from netCDF, CDL, PP and
  UM datasets,

* create new field and domain constructs in memory,

* inspect field and domain constructs,

* test whether two constructs are the same,

* modify field and domain construct metadata and data,

* create subspaces of field and domain constructs,

* write and append field constructs to netCDF datasets on disk,

* incorporate, and create, metadata stored in external files (*new in
  version 3.0.0*),

* read, write, and create data that have been compressed by convention
  (i.e. ragged or gathered arrays), whilst presenting a view of the
  data in its uncompressed form,

* read, write, and create coordinates defined by geometry cells,

* read netCDF and CDL datasets containing hierarchical groups,

* combine field constructs arithmetically,

* manipulate field construct data by arithmetical and trigonometrical
  operations,

* perform statistical collapses on field constructs,

* perform histogram, percentile and binning operations on field
  constructs,

* regrid field constructs with (multi-)linear, nearest neighbour,
  first- and second-order conservative and higher order patch recovery
  methods,

* apply convolution filters to field constructs,

* calculate derivatives of field constructs,

* create field constructs to create derived quantities (such as
  vorticity).

All of the above use LAMA functionality, which allows multiple fields
larger than the available memory to exist and be manipulated. (Note:
work is underway to replace this functionality with a `dask`
implementation.)


Visualization
=============

Powerful, flexible, and very simple to produce visualizations of field
constructs are available with the `cfplot` package
(http://ajheaps.github.io/cf-plot), that needs to be installed
seprately to the `cf` package.

See the `cfplot` gallery (http://ajheaps.github.io/cf-plot/gallery.html)
for the full range range plotting possibilities with example code.


Command line utilities
======================

During installation the ``cfa`` command line utility is also
installed, which

* generates text descriptions of field constructs contained in files,
  and

* creates new datasets aggregated from existing files.


Tests
=====

Tests are run from within the ``cf/test`` directory:

    python run_tests.py
"""

# Get dependencies
requirements = open("requirements.txt", "r")
install_requires = requirements.read().splitlines()

tests_require = (
    [
        "pytest",
        "pycodestyle",
        "coverage",
    ],
)
extras_require = {
    "required C libraries": ["udunits2==2.2.25"],
    "regridding": ["esmpy", "ESMF>=8.0"],
    "convolution filters, derivatives, relative vorticity": ["scipy>=1.1.0"],
    "subspacing with multi-dimensional construct cells": ["matplotlib>=3.0.0"],
    "parallel processing": ["mpich", "mpi4py>=3.0.0"],
    "documentation": [
        "sphinx>=2,<=4",
        "sphinx-copybutton",
        "sphinx-toggleprompt",
        "sphinxcontrib-spelling",
    ],
    "pre-commit hooks": [
        "pre-commit",
        "black",
        "docformatter",
        "flake8",
    ],
}

setup(
    name="cf-python",
    long_description=long_description,
    version=version,
    description="A CF-compliant earth science data analysis library",
    author="David Hassell",
    maintainer="David Hassell",
    maintainer_email="david.hassell@ncas.ac.uk",
    author_email="david.hassell@ncas.ac.uk",
    url="https://ncas-cms.github.io/cf-python",
    platforms=["Linux", "MacOS"],
    license="MIT",
    keywords=[
        "cf",
        "netcdf",
        "UM",
        "data",
        "science",
        "oceanography",
        "meteorology",
        "climate",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Utilities",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=[
        "cf",
        "cf.mixin",
        "cf.data",
        "cf.data.array.abstract",
        "cf.docstring",
        "cf.read_write",
        "cf.read_write.um",
        "cf.read_write.netcdf",
        "cf.regrid",
        "cf.umread_lib",
        "cf.test",
    ],
    package_data={"cf": package_data},
    scripts=["scripts/cfa"],
    python_requires=">=3.7",
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    # install_requires=[
    #     'netCDF4>=1.5.3',
    #     'cftime>=1.1.3',
    #     'numpy>=1.15',
    #     'cfdm>=1.8.5, <1.9',
    #     'psutil>=0.6.0',
    #     'cfunits>=3.2.7'
    #     'scipy>=1.1.0',
    #     'matplotlib>=3.0.0',
    #     'mpi4py>=3.0.0',
    #     'ESMF>=8.0',
    #     'udunits2==2.2.25',
    # ],
    #
    # https://docs.python.org/2/distutils/apiref.html:
    cmdclass={"build": build_umread},
)



import datetime
import faulthandler
import inspect
import os
import tempfile
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class ExternalVariableTest(unittest.TestCase):
    def setUp(self):
        self.parent_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "parent.nc"
        )
        self.external_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "external.nc"
        )
        self.combined_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "combined.nc"
        )
        self.external_missing_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "external_missing.nc"
        )

        self.test_only = []

        (fd, self.tempfilename) = tempfile.mkstemp(
            suffix=".nc", prefix="cf_", dir="."
        )
        os.close(fd)
        (fd, self.tempfilename_parent) = tempfile.mkstemp(
            suffix=".nc", prefix="cf_parent_", dir="."
        )
        os.close(fd)
        (fd, self.tempfilename_external) = tempfile.mkstemp(
            suffix=".nc", prefix="cf_external_", dir="."
        )
        os.close(fd)

    def tearDown(self):
        os.remove(self.tempfilename)
        os.remove(self.tempfilename_parent)
        os.remove(self.tempfilename_external)

    def test_EXTERNAL_READ(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Read the parent file on its own, without the external file
        f = cf.read(self.parent_file, verbose=0)

        self.assertEqual(len(f), 1)

        for i in f:
            repr(i)
            str(i)
            i.dump(display=False)

        f = f[0]

        cell_measure = f.constructs.filter_by_identity("measure:area").value()

        self.assertTrue(cell_measure.nc_get_external())
        self.assertEqual(cell_measure.nc_get_variable(), "areacella")
        self.assertEqual(cell_measure.properties(), {})
        self.assertFalse(cell_measure.has_data())

        # External file contains only the cell measure variable
        f = cf.read(self.parent_file, external=[self.external_file], verbose=0)

        c = cf.read(self.combined_file, verbose=0)

        for i in c + f:
            repr(i)
            str(i)
            i.dump(display=False)

        cell_measure = (
            f[0].constructs.filter_by_identity("measure:area").value()
        )

        self.assertEqual(len(f), 1)
        self.assertEqual(len(c), 1)

        for i in range(len(f)):
            self.assertTrue(c[i].equals(f[i], verbose=2))

        # External file contains other variables
        f = cf.read(self.parent_file, external=self.combined_file, verbose=0)

        for i in f:
            repr(i)
            str(i)
            i.dump(display=False)

        self.assertEqual(len(f), 1)
        self.assertEqual(len(c), 1)

        for i in range(len(f)):
            self.assertTrue(c[i].equals(f[i], verbose=2))

        # Two external files
        f = cf.read(
            self.parent_file,
            external=[self.external_file, self.external_missing_file],
            verbose=0,
        )

        for i in f:
            repr(i)
            str(i)
            i.dump(display=False)

        self.assertEqual(len(f), 1)
        self.assertEqual(len(c), 1)

        for i in range(len(f)):
            self.assertTrue(c[i].equals(f[i], verbose=2))

    def test_EXTERNAL_WRITE(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        parent = cf.read(self.parent_file)
        combined = cf.read(self.combined_file)

        # External file contains only the cell measure variable
        f = cf.read(self.parent_file, external=self.external_file)

        cf.write(f, self.tempfilename)
        g = cf.read(self.tempfilename)

        self.assertEqual(len(g), len(combined))

        for i in range(len(g)):
            self.assertTrue(combined[i].equals(g[i], verbose=2))

        cell_measure = g[0].constructs("measure:area").value()

        self.assertFalse(cell_measure.nc_get_external())
        cell_measure.nc_set_external(True)
        self.assertTrue(cell_measure.nc_get_external())
        self.assertTrue(cell_measure.properties())
        self.assertTrue(cell_measure.has_data())

        self.assertTrue(
            g[0]
            .constructs.filter_by_identity("measure:area")
            .value()
            .nc_get_external()
        )

        cf.write(
            g,
            self.tempfilename_parent,
            external=self.tempfilename_external,
            verbose=0,
        )

        h = cf.read(self.tempfilename_parent, verbose=0)

        self.assertEqual(len(h), len(parent))

        for i in range(len(h)):
            self.assertTrue(parent[i].equals(h[i], verbose=2))

        h = cf.read(self.tempfilename_external)
        external = cf.read(self.external_file)

        self.assertEqual(len(h), len(external))

        for i in range(len(h)):
            self.assertTrue(external[i].equals(h[i], verbose=2))

    def test_EXTERNAL_AGGREGATE(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Read parent file without the external file, taking first field to test
        f = cf.read(self.parent_file, verbose=0)[0]
        measure_name = "measure:area"

        # Split f into parts (take longitude with 3 x 3 = 9 points) so can
        # test the difference between the aggregated result and original f.
        # Note all parts retain the external variable cell measure.
        f_lon_thirds = [f[:, :3], f[:, 3:6], f[:, 6:]]

        g = cf.aggregate(f_lon_thirds)

        self.assertEqual(len(g), 1)

        # Check cell measure construct from external variable has been removed
        self.assertFalse(g[0].cell_measures())

        # Check aggregated field is identical to original with measure removed
        f0 = f.copy()
        f0.del_construct(measure_name)
        self.assertEqual(g[0], f0)

        # Also check aggregation did not remove the measure from the inputs
        for part in f_lon_thirds:
            cell_measure = part.constructs.filter_by_identity(
                "measure:area"
            ).value()
            self.assertTrue(cell_measure.nc_get_external())

        # Now try aggregating when one part doesn't have the cell measure,
        # expecting all of the parts to still aggregate back to one field
        # without the external measure (rather than  2->1 => 1 + 1 = 2 fields).
        f_lon_thirds[1].del_construct(measure_name)
        g = cf.aggregate(f_lon_thirds)
        self.assertEqual(len(g), 1)
        self.assertFalse(g[0].cell_measures())

        # Also check measure was not removed from, or added to, any input
        for part in [f_lon_thirds[0], f_lon_thirds[2]]:
            cm = part.constructs.filter_by_identity("measure:area").value()
            self.assertTrue(cm.nc_get_external())
        self.assertFalse(f_lon_thirds[1].cell_measures())


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import cfdm


class Datum(cfdm.Datum):
    """A datum component of a coordinate reference construct of the CF
    data model.

    A datum is a complete or partial definition of the zeroes of the
    dimension and auxiliary coordinate constructs which define a
    coordinate system.

    The datum may contain the definition of a geophysical surface
    which corresponds to the zero of a vertical coordinate construct,
    and this may be required for both horizontal and vertical
    coordinate systems.

    Elements of the datum not specified may be implied by the
    properties of the dimension and auxiliary coordinate constructs
    referenced by the `cf.CoordinateReference` instance that contains
    the datum.

    .. versionadded:: 3.0.0

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)



"""This class is not in the cf.mixin package because it needs to be
imported by cf.Data, and some of the other mixin classes in cf.mixin
themsleves import cf.Data, which would lead to a circular import
situation."""
from .docstring import _docstring_substitution_definitions


class Container:
    """Mixin class for storing components.

    .. versionadded:: 3.7.0

    """

    def __docstring_substitutions__(self):
        """Define docstring substitutions that apply to this class and
        all of its subclasses.

        These are in addtion to, and take precendence over, docstring
        substitutions defined by the base classes of this class.

        See `_docstring_substitutions` for details.

        .. versionadded:: 3.7.0

        .. seealso:: `_docstring_substitutions`

        :Returns:

            `dict`
                The docstring substitutions that have been applied.

        """
        return _docstring_substitution_definitions

    def __docstring_package_depth__(self):
        """Return the package depth for {{package}} docstring
        substitutions.

        See `_docstring_package_depth` for details.

        """
        return 0



import datetime
import faulthandler
import os
import unittest

import numpy

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class MathTest(unittest.TestCase):
    filename1 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "regrid_file1.nc"
    )

    def test_relative_vorticity_distance(self):
        x_min = 0.0
        x_max = 100.0
        dx = 1.0

        x_1d = numpy.arange(x_min, x_max, dx)
        size = x_1d.size

        data_1d = x_1d * 2.0 + 1.0
        data_2d = numpy.broadcast_to(data_1d[numpy.newaxis, :], (size, size))

        dim_x = cf.DimensionCoordinate(
            data=cf.Data(x_1d, "m"), properties={"axis": "X"}
        )
        dim_y = cf.DimensionCoordinate(
            data=cf.Data(x_1d, "m"), properties={"axis": "Y"}
        )

        u = cf.Field()
        X = u.set_construct(cf.DomainAxis(size=dim_x.data.size))
        Y = u.set_construct(cf.DomainAxis(size=dim_y.data.size))
        u.set_construct(dim_x, axes=[X])
        u.set_construct(dim_y, axes=[Y])
        u.set_data(cf.Data(data_2d, "m/s"), axes=("Y", "X"))

        v = cf.Field()
        v.set_construct(cf.DomainAxis(size=dim_x.data.size))
        v.set_construct(cf.DomainAxis(size=dim_y.data.size))
        v.set_construct(dim_x, axes=[X])
        v.set_construct(dim_y, axes=[Y])
        v.set_data(cf.Data(data_2d, "m/s"), axes=("X", "Y"))

        rv = cf.relative_vorticity(u, v, one_sided_at_boundary=True)
        self.assertTrue((rv.array == 0.0).all())

    def test_relative_vorticity_latlong(self):
        lat_min = -90.0
        lat_max = 90.0
        dlat = 1.0

        lat_1d = numpy.arange(lat_min, lat_max, dlat)
        lat_size = lat_1d.size

        lon_min = 0.0
        lon_max = 359.0
        dlon = 1.0

        lon_1d = numpy.arange(lon_min, lon_max, dlon)
        lon_size = lon_1d.size

        u_1d = lat_1d * 2.0 + 1.0
        u_2d = numpy.broadcast_to(u_1d[numpy.newaxis, :], (lon_size, lat_size))

        v_1d = lon_1d * 2.0 + 1.0
        v_2d = numpy.broadcast_to(v_1d[:, numpy.newaxis], (lon_size, lat_size))
        v_2d = v_2d * numpy.cos(lat_1d * numpy.pi / 180.0)[numpy.newaxis, :]

        rv_array = (
            u_2d
            / cf.Data(6371229.0, "meters")
            * numpy.tan(lat_1d * numpy.pi / 180.0)[numpy.newaxis, :]
        )

        dim_x = cf.DimensionCoordinate(
            data=cf.Data(lon_1d, "degrees_east"), properties={"axis": "X"}
        )
        dim_y = cf.DimensionCoordinate(
            data=cf.Data(lat_1d, "degrees_north"), properties={"axis": "Y"}
        )

        u = cf.Field()
        u.set_construct(cf.DomainAxis(size=lon_1d.size))
        u.set_construct(cf.DomainAxis(size=lat_1d.size))
        u.set_construct(dim_x)
        u.set_construct(dim_y)
        u.set_data(cf.Data(u_2d, "m/s"), axes=("X", "Y"))
        u.cyclic("X", period=360.0)

        v = cf.Field()
        v.set_construct(cf.DomainAxis(size=lon_1d.size))
        v.set_construct(cf.DomainAxis(size=lat_1d.size))
        v.set_construct(dim_x)
        v.set_construct(dim_y)
        v.set_data(cf.Data(v_2d, "m/s"), axes=("X", "Y"))
        v.cyclic("X", period=360.0)

        rv = cf.relative_vorticity(u, v, wrap=False)
        self.assertTrue(numpy.allclose(rv.array, rv_array))

    def test_curl_xy(self):
        f = cf.example_field(0)

        # Spherical polar coordinates
        theta = 90 - f.convert("Y", full_domain=True)
        sin_theta = theta.sin()

        radius = 2
        r = f.radius(radius)

        for wrap in (False, True, None):
            for one_sided in (True, False):
                x, y = f.grad_xy(
                    radius=radius, x_wrap=wrap, one_sided_at_boundary=one_sided
                )

                c = cf.curl_xy(
                    x,
                    y,
                    radius=radius,
                    x_wrap=wrap,
                    one_sided_at_boundary=one_sided,
                )

                self.assertTrue(c.Units == cf.Units("m-2 rad-2"))

                term1 = (x * sin_theta).derivative(
                    "Y", one_sided_at_boundary=one_sided
                )
                term2 = y.derivative(
                    "X", wrap=wrap, one_sided_at_boundary=one_sided
                )

                c0 = (term1 - term2) / (sin_theta * r)

                # Check the data
                with cf.rtol(1e-10):
                    self.assertTrue((c.data == c0.data).all())

                del c.long_name
                c0.set_data(c.data)
                self.assertTrue(c.equals(c0))

        # Cartesian coordinates
        dim_x = f.dimension_coordinate("X")
        dim_y = f.dimension_coordinate("Y")
        dim_x.override_units("m", inplace=True)
        dim_y.override_units("m", inplace=True)
        dim_x.standard_name = "projection_x_coordinate"
        dim_y.standard_name = "projection_y_coordinate"
        f.cyclic("X", iscyclic=False)

        for wrap in (False, True, None):
            for one_sided in (True, False):
                x, y = f.grad_xy(x_wrap=wrap, one_sided_at_boundary=one_sided)

                d = cf.div_xy(
                    x, y, x_wrap=wrap, one_sided_at_boundary=one_sided
                )

                self.assertTrue(d.Units == cf.Units("m-2"))

                term1 = x.derivative(
                    "X", wrap=wrap, one_sided_at_boundary=one_sided
                )
                term2 = y.derivative("Y", one_sided_at_boundary=one_sided)

                d0 = term1 + term2

                del d.long_name
                del d0.long_name

                self.assertTrue(d.equals(d0, rtol=1e-10))

    def test_div_xy(self):
        f = cf.example_field(0)

        # Spherical polar coordinates
        theta = 90 - f.convert("Y", full_domain=True)
        sin_theta = theta.sin()

        radius = 2
        r = f.radius(radius)

        for wrap in (False, True, None):
            for one_sided in (False, True):
                x, y = f.grad_xy(
                    radius=radius, x_wrap=wrap, one_sided_at_boundary=one_sided
                )

                d = cf.div_xy(
                    x,
                    y,
                    radius=radius,
                    x_wrap=wrap,
                    one_sided_at_boundary=one_sided,
                )

                self.assertTrue(d.Units == cf.Units("m-2 rad-2"), d.Units)

                term1 = x.derivative(
                    "X", wrap=wrap, one_sided_at_boundary=one_sided
                )
                term2 = (y * sin_theta).derivative(
                    "Y", one_sided_at_boundary=one_sided
                )

                d0 = (term1 + term2) / (sin_theta * r)

                # Check the data
                with cf.rtol(1e-10):
                    self.assertTrue((d.data == d0.data).all())

                del d.long_name
                d0.set_data(d.data)
                self.assertTrue(d.equals(d0))

        # Cartesian coordinates
        dim_x = f.dimension_coordinate("X")
        dim_y = f.dimension_coordinate("Y")
        dim_x.override_units("m", inplace=True)
        dim_y.override_units("m", inplace=True)
        dim_x.standard_name = "projection_x_coordinate"
        dim_y.standard_name = "projection_y_coordinate"
        f.cyclic("X", iscyclic=False)

        for wrap in (False, True, None):
            for one_sided in (True, False):
                x, y = f.grad_xy(x_wrap=wrap, one_sided_at_boundary=one_sided)

                d = cf.div_xy(
                    x, y, x_wrap=wrap, one_sided_at_boundary=one_sided
                )

                self.assertTrue(d.Units == cf.Units("m-2"))

                term1 = x.derivative(
                    "X", wrap=wrap, one_sided_at_boundary=one_sided
                )
                term2 = y.derivative("Y", one_sided_at_boundary=one_sided)

                d0 = term1 + term2

                del d.long_name
                del d0.long_name
                self.assertTrue(d.equals(d0, rtol=1e-10))

    def test_differential_operators(self):
        f = cf.example_field(0)

        radius = 2

        fx, fy = f.grad_xy(radius=radius, one_sided_at_boundary=True)
        c = cf.curl_xy(fx, fy, radius=radius, one_sided_at_boundary=True)

        # Divergence of curl is zero
        dc = cf.div_xy(c, c, radius=radius, one_sided_at_boundary=True)

        zeros = dc.copy()
        zeros[...] = 0
        self.assertTrue(dc.data.equals(zeros.data, rtol=0, atol=1e-15))

        # Curl of gradient is zero
        cg = cf.curl_xy(fx, fy, radius=radius, one_sided_at_boundary=True)

        zeros = cg.copy()
        zeros[...] = 0
        self.assertTrue(cg.data.equals(zeros.data, rtol=0, atol=1e-15))


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



from . import ConstructList, mixin


class DomainList(mixin.FieldDomainList, ConstructList):
    """An ordered sequence of domain constructs.

    A domain list supports the python list-like operations (such as
    indexing and methods like `!append`). These methods provide
    functionality similar to that of a built-in list. The main
    difference is that when a domain construct element needs to be
    assesed for equality its `~cf.Domain.equals` method is used,
    rather than the ``==`` operator.

    .. versionadded:: 3.11.0

    """

    def __init__(self, domains=None):
        """Initialisation.

        :Parameters:

            domains: (sequence of) `Domain`, optional
                 Create a new list with these domain constructs.

        """
        super().__init__(constructs=domains)



from .. import _found_ESMF

if _found_ESMF:
    try:
        import ESMF
    except Exception:
        _found_ESMF = False

if _found_ESMF:
    regrid_method_map = {
        "linear": ESMF.RegridMethod.BILINEAR,  # see comment below...
        "bilinear": ESMF.RegridMethod.BILINEAR,  # (for back compat)
        "conservative": ESMF.RegridMethod.CONSERVE,
        "conservative_1st": ESMF.RegridMethod.CONSERVE,
        "conservative_2nd": ESMF.RegridMethod.CONSERVE_2ND,
        "nearest_dtos": ESMF.RegridMethod.NEAREST_DTOS,
        "nearest_stod": ESMF.RegridMethod.NEAREST_STOD,
        "patch": ESMF.RegridMethod.PATCH,
    }
    # ... diverge from ESMF with respect to name for bilinear method
    # by using 'linear' because 'bi' implies 2D linear interpolation,
    # which could mislead or confuse for Cartesian regridding in 1D or
    # 3D.
else:
    # Initialize a place holder that may be imported when ESMF is not
    # installed
    regrid_method_map = {}

regrid_method_map_inverse = {v: k for k, v in regrid_method_map.items()}

conservative_regridding_methods = (
    "conservative",
    "conservative_1st",
    "conservative_2nd",
)

regridding_methods = (
    "linear",  # prefer over 'bilinear' as of v3.2.0
    "bilinear",  # only for backward compatibility, use & document 'linear'
    "patch",
    "nearest_stod",
    "nearest_dtos",
) + conservative_regridding_methods


class RegridOperator:
    """A regridding operator between two fields.

    Stores an `ESMF.Regrid` regridding operator and associated
    parameters that describe the complete coordinate system of the
    destination grid.

    """

    def __init__(self, regrid, name, **parameters):
        """**Initialization**

        :Parameters:

            regrid: `ESMF.Regrid`
                The `ESMF` regridding operator between two fields.

            name: `str`
                A name that defines the context of the destination
                grid parameters.

            parameters: `dict`
               Parameters that describe the complete coordinate system
               of the destination grid.

               Any parameter names and values are allowed, and it is
               assumed that the these are well defined during the
               creation and subsequent use of a `RegridOperator`
               instance.

        """
        self._regrid = regrid
        self._name = name
        self._parameters = parameters

    def __del__(self):
        """Calls the `ESMF.Regrid` destroy method."""
        self._regrid.destroy()

    def __repr__(self):
        return (
            f"<CF {self.__class__.__name__}: "
            f"{self._name}, method={self.method}>"
        )

    @property
    def method(self):
        """The regridding method.

        **Examples**

        >>> r.method
        'conservative'

        """
        method = regrid_method_map_inverse[self._regrid.regrid_method]
        if method == "bilinear":
            method = "linear"

        return method

    @property
    def name(self):
        """The name of the regrid method.

        **Examples**

        >>> r.name
        'regrids'

        """
        return self._name

    @property
    def parameters(self):
        """The parameters that describe the destination grid.

        Any parameter names and values are allowed, and it is assumed
        that the these are well defined during the creation and
        subsequent use of a `RegridOperator` instance.

        **Examples**

        >>> type(r.parameters)
        dict

        """
        return self._parameters

    @property
    def regrid(self):
        """The contained regridding operator.

        **Examples**

        >>> type(r.regrid)
        ESMF.api.regrid.Regrid

        """
        return self._regrid

    def check_method(self, method):
        """Whether the given method is equivalent to the regridding
        method.

        :Parameters:

            method: `str`
                A regridding method, such as ``'conservative'``.

        :Returns:

            `bool`
                Whether or not method is equivalent to the regridding
                method.

        **Examples**

        >>> r.method
        'conservative'
        >>> r.check_method('conservative')
        True
        >>> r.check_method('conservative_1st')
        True
        >>> r.check_method('conservative_2nd')
        False

        """
        return regrid_method_map.get(method) == self._regrid.regrid_method

    def copy(self):
        """Return a copy.

        The contained `ESMF.Regrid` instance (see `regrid`) is shallow
        copied and the "dst" parameter, which is a `Field` or `dict`
        instance, is copied with its `!copy` method.

        :Returns:

            `RegridOperator`
                The copy.

        **Examples**

        >>> s = r.copy()

        """
        parameters = self._parameters
        if "dst" in parameters:
            parameters = parameters.copy()
            parameters["dst"] = parameters["dst"].copy()

        return type(self)(
            regrid=self._regrid.copy(), name=self._name, **parameters
        )

    def destroy(self):
        """Free the memory allocated by the `ESMF.Regrid` instance.

        **Examples**

        >>> r.destroy()

        """
        # explicitly release memory for source ESMF objects
        self._regrid.srcfield.grid.destroy()
        self._regrid.srcfield.destroy()
        self._regrid.src_frac_field.destroy()

        # explicitly release memory for destination ESMF objects
        self._regrid.dstfield.grid.destroy()
        self._regrid.dstfield.destroy()
        self._regrid.dst_frac_field.destroy()

        # explicitly release memory for ESMF Regrid
        self._regrid.destroy()

    def get_parameter(self, parameter):
        """Return a regrid operation parameter.

        **Examples**

        >>> r.get_parameter('ignore_degenerate')
        True
        >>> r.get_parameter('x')
        Traceback
            ...
        ValueError: RegridOperator has no 'x' parameter

        """
        try:
            return self._parameters[parameter]
        except KeyError:
            raise ValueError(
                f"{self.__class__.__name__} has no {parameter!r} parameter"
            )



import cfdm

from .mixin import ArrayMixin, RaggedArrayMixin


class RaggedContiguousArray(
    RaggedArrayMixin, ArrayMixin, cfdm.RaggedContiguousArray
):
    """An underlying contiguous ragged array.

    A collection of features stored using a contiguous ragged array
    combines all features along a single dimension (the "sample
    dimension") such that each feature in the collection occupies a
    contiguous block.

    The information needed to uncompress the data is stored in a
    "count variable" that gives the size of each block.

    It is assumed that the compressed dimension is the left-most
    dimension in the compressed array.

    See CF section 9 "Discrete Sampling Geometries".

    .. versionadded:: 3.0.0

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        .. versionadded:: 3.0.0

        """
        return super().__repr__().replace("<", "<CF ", 1)



from .regridoperator import (
    RegridOperator,
    conservative_regridding_methods,
    regrid_method_map,
    regrid_method_map_inverse,
    regridding_methods,
)
from .utils import (
    create_Grid,
    create_Field,
    create_Regrid,
    get_cartesian_coords,
    grids_have_same_coords,
    grids_have_same_masks,
    regrid_get_latlon,
    regrid_get_axis_indices,
    regrid_get_coord_order,
    regrid_get_section_shape,
    regrid_check_bounds,
    regrid_check_method,
    regrid_check_use_src_mask,
    regrid_get_reordered_sections,
    regrid_get_destination_mask,
    regrid_fill_fields,
    regrid_compute_field_mass,
    regrid_get_regridded_data,
    regrid_update_coordinate_references,
    regrid_copy_coordinate_references,
    regrid_use_bounds,
    regrid_update_coordinates,
    regrid_initialize,
    regrid_get_operator_method,
    regrid_create_operator,
    run_Regrid,
)



import logging
from copy import deepcopy
from operator import __and__ as operator_and
from operator import __or__ as operator_or

from .data import Data
from .decorators import (
    _deprecated_kwarg_check,
    _display_or_return,
    _manage_log_level_via_verbosity,
)
from .functions import (
    _DEPRECATION_ERROR_ATTRIBUTE,
    _DEPRECATION_ERROR_FUNCTION,
    _DEPRECATION_ERROR_FUNCTION_KWARGS,
)
from .functions import equals as _equals
from .functions import inspect as _inspect
from .units import Units

logger = logging.getLogger(__name__)

# Alias for builtin set, since there is a 'set' function
builtin_set = set


class Query:
    '''Encapsulate a condition for subsequent evaluation.

    A condition that may be applied to any object may be stored in a
    `Query` object. A `Query` object encapsulates a condition, such as
    "strictly less than 3". When applied to an object, via its
    `evaluate` method or the Python `==` operator, the condition is
    evaluated in the context of that object.

       >>> c = cf.Query('lt', 3)
       >>> c
       <CF Query: (lt 3)>
       >>> c.evaluate(2)
       True
       >>> c == 2
       True
       >>> c != 2
       False
       >>> c.evaluate(3)
       False
       >>> c == cf.Data([1, 2, 3])
       <CF Data(3): [True, True, False]>
       >>> c == numpy.array([1, 2, 3])
       array([True, True, False])

    The following operators are supported when constructing `Query`
    instances:

    =========  ===================================
    Operator   Description
    =========  ===================================
    ``'lt'``   A "strictly less than" condition
    ``'le'``   A "less than or equal" condition
    ``'gt'``   A "strictly greater than" condition
    ``'ge'``   A "greater than or equal" condition
    ``'eq'``   An "equal" condition
    ``'ne'``   A "not equal" condition
    ``'wi'``   A "within a range" condition
    ``'wo'``   A "without a range" condition
    ``'set'``  A "member of set" condition
    =========  ===================================

    **Compound queries**

    Multiple conditions may be combined with the Python bitwise "and"
    (`&`) and "or" (`|`) operators to form a new `Query` object.

       >>> ge3 = cf.Query('ge', 3)
       >>> lt5 = cf.Query('lt', 5)
       >>> c = ge3 & lt5
       >>> c
       <CF Query: [(ge 3) & (lt 5)]>
       >>> c == 2
       False
       >>> c != 2
       True
       >>> c = ge3 | lt5
       >>> c
       <CF Query: [(ge 3) | (lt 5)]>
       >>> c == 2
       True
       >>> c &= cf.Query('set', [1, 3, 5])
       >>> c
       <CF Query: [[(ge 3) | (lt 5)] & (set [1, 3, 5])]>
       >>> c == 2
       False
       >>> c == 3
       True

    A condition can be applied to an attribute of an object.

       >>> upper_bounds_ge_minus4 = cf.Query('ge', -4, attr='upper_bounds')
       >>> x
       <CF DimensionCoordinate: grid_longitude(9) degrees>
       >>> print(x.bounds.array)
       [[-4.92 -4.48]
        [-4.48 -4.04]
        [-4.04 -3.6 ]
        [-3.6  -3.16]
        [-3.16 -2.72]
        [-2.72 -2.28]
        [-2.28 -1.84]
        [-1.84 -1.4 ]
        [-1.4  -0.96]]
       >>> print((upper_bounds_ge_minus4 == x).array)
       [False False  True  True  True  True  True  True  True]
       >>> upper_bounds_ge_minus4 = cf.Query('ge', -4, attr='upper_bounds')

    A condition can also be applied to attributes of attributes of an
    object.

       >>> t
       <CF DimensionCoordinate: time(4) >
       >>> t.lower_bounds.month.array
       array([12,  3,  6,  9])
       >>> c = cf.Query('ge', 8, attr='lower_bounds.month')
       >>> c == t
       <CF Data(4): [True, ..., True]>
       >>> (c == t).array
       array([ True,  False, False, True])


    **The query interface**

    In general, the query operator must be permitted between the value
    of the condition and the operand for which it is being
    evaluated. For example, when the value is an `int`, the query
    works if the operand is also an `int`, but fails if it is a
    `list`:

       >>> c = cf.Query('lt', 2)
       >>> c == 1
       True
       >>> c == [1, 2, 3]
       TypeError: '<' not supported between instances of 'list' and 'int'

    This behaviour is overridden if the operand has an appropriate
    "query interface" method. When such a method exists, it is used
    instead of the equivalent built-in Python operator.

    ======================  ==============================================
    Query interface method  Description
    ======================  ==============================================
    `__query_lt__`          Called when a ``'lt'`` condition is evaluated
    `__query_le__`          Called when a ``'le'`` condition is evaluated
    `__query_gt__`          Called when a ``'gt'`` condition is evaluated
    `__query_ge__`          Called when a ``'ge'`` condition is evaluated
    `__query_eq__`          Called when an ``'eq'`` condition is evaluated
    `__query_ne__`          Called when a ``'ne'`` condition is evaluated
    `__query_wi__`          Called when a ``'wi'`` condition is evaluated
    `__query_wo__`          Called when a ``'wo'`` condition is evaluated
    `__query_set__`         Called when a ``'set'`` condition is evaluated
    ======================  ==============================================

    In all cases the query value is the only, mandatory argument of
    the method.

       >>> class myList(list):
       ...     pass
       ...
       >>> class myList_with_override(list):
       ...     def __query_lt__(self, value):
       ...         """Apply the < operator element-wise"""
       ...         return type(self)([x < value for x in self])
       ...
       >>> c == myList([1, 2, 3])
       TypeError: '<' not supported between instances of 'myList' and 'int'
       >>> c == myList_with_override([1, 2, 3])
       [True, False, False]

    When the condition is on an attribute, or nested attributes, of
    the operand, the query interface method is looked for on the
    attribute object, rather than the parent object.

    If the value has units then the argument passed to query interface
    method is automatically a `Data` object that attaches the units to
    the value.

    '''

    isquery = True

    @_deprecated_kwarg_check("exact", version="3.0.0", removed_at="4.0.0")
    def __init__(self, operator, value, units=None, attr=None, exact=True):
        """**Initialisation**

        :Parameters:

            operator: `str`
                The query operator.

            value:
                The value of the condition.

            units: `str` or `Units`, optional
                The units of *value*. By default, the same units as
                the operand being tested are assumed, if
                applicable. If *units* is specified and *value*
                already has units (such as those attached to a `Data`
                object), then the pair of units must be equivalent.

            attr: `str`, optional
                Apply the condition to the attribute, or nested
                attributes, of the operand, rather than the operand
                itself. Nested attributes are specified by separating
                them with a ``.``. For example, the "month" attribute
                of the "bounds" attribute is specified as
                ``'bounds.month'``. See also the `addattr` method.

            exact: deprecated at version 3.0.0.
                Use `re.compile` objects in *value* instead.

        """
        if units is not None:
            value_units = getattr(value, "Units", None)
            if value_units is None:
                value = Data(value, units)
            elif not value_units.equivalent(Units(units)):
                raise ValueError(
                    f"'{value_units}' and '{Units(units)}' are not equivalent "
                    f"units therefore the query does not make physical "
                    "sense."
                )

        self._operator = operator
        self._value = value
        self._compound = False

        if attr:
            self._attr = tuple(attr.split("."))
        else:
            self._attr = ()

        self._bitwise_operator = None

        self.query_type = operator

        self._NotImplemented_RHS_Data_op = True

    def __deepcopy__(self, memo):
        """Used if copy.deepcopy is called on the variable."""
        return self.copy()

    def __eq__(self, x):
        """The rich comparison operator ``==``

        x.__eq__(y) <==> x==y

        x.__eq__(y) <==> x.evaluate(y)

        """
        return self._evaluate(x, ())

    def __ne__(self, x):
        """The rich comparison operator ``!=``

        x.__ne__(y) <==> x!=y

        x.__ne__(y) <==> (x==y)==False

        """
        # Note that it is important to use the == operator
        return self._evaluate(x, ()) == False  # ignore PEP8 E712 due to above

    def __and__(self, other):
        """The binary bitwise operation ``&``

        Combine two queries with a logical And operation. If the
        `!value` of both queries is the same then it will be retained on
        the compound query.

        x.__and__(y) <==> x&y

        """
        Q = type(self)
        new = Q.__new__(Q)

        new._operator = None
        new._compound = (self.copy(), other.copy())
        new._bitwise_operator = operator_and
        new._attr = ()

        # If the value of the two queries is the same then retain it
        # on the compound query
        value0 = self._value
        value1 = other._value
        if value0 is None or value1 is None or value0 != value1:
            new._value = None
        else:
            new._value = deepcopy(value0)

        new._NotImplemented_RHS_Data_op = True

        return new

    def __iand__(self, other):
        """The augmented bitwise assignment ``&=``

        x.__iand__(y) <==> x&=y

        """
        return self & other

    def __or__(self, other):
        """The binary bitwise operation ``|``

        Combine two queries with a logical Or operation. If the
        `!value` of both queries is the same then it will be retained on
        the compound query.

        x.__or__(y) <==> x|y

        """
        Q = type(self)
        new = Q.__new__(Q)

        new._operator = None
        new._compound = (self, other)
        new._bitwise_operator = operator_or
        new._attr = ()

        # If the value of the two queries is the same then retain it
        # on the compound query
        value0 = self._value
        value1 = other._value
        if value0 is None or value1 is None or value0 != value1:
            new._value = None
        else:
            new._value = deepcopy(value0)

        new._NotImplemented_RHS_Data_op = True

        return new

    def __ior__(self, other):
        """The augmented bitwise assignment ``|=``

        x.__ior__(y) <==> x|=y

        """
        return self | other

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return f"<CF {self.__class__.__name__}: {self}>"

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        """
        attr = ".".join(self._attr)

        if not self._compound:
            out = f"{attr}({self._operator} {self._value!s})"
        else:
            bitwise_operator = repr(self._bitwise_operator)
            if "and_" in bitwise_operator:
                bitwise_operator = "&"
            elif "or_" in bitwise_operator:
                bitwise_operator = "|"

            out = (
                f"{attr}[{self._compound[0]} {bitwise_operator} "
                f"{self._compound[1]}]"
            )

        return out

    @property
    def attr(self):
        """The object attribute on which to apply the query condition.

        **Examples**

        >>> q = cf.Query('ge', 4)
        >>> print(q.attr)
        ()
        >>> q = cf.Query('le', 6, attr='year')
        >>> q.attr
        ('year',)

        """
        return self._attr

    @property
    def operator(self):
        """The query operator.

        Compound conditions return `None`.

        **Examples**

        >>> q = cf.Query('ge', 4)
        >>> q.operator
        'ge'
        >>> q |= cf.Query('le', 6)
        >>> print(q.operator)
        None

        """
        return self._operator

    @property
    def value(self):
        """The value of the condition encapsulated by the query.

        An exception is raised for compound conditions.

        **Examples**

        >>> q = cf.Query('ge', 4)
        >>> q.value
        4
        >>> q |= cf.Query('le', 6)
        >>> q.value
        AttributeError: Compound query doesn't have attribute 'value'

        """
        value = self._value
        if value is None:
            raise AttributeError(
                "Compound query doesn't have attribute 'value'"
            )

        return value

    def addattr(self, attr):
        """Return a `Query` object with a new left hand side operand
        attribute to be used during evaluation. TODO.

        If another attribute has previously been specified, then the new
        attribute is considered to be an attribute of the existing
        attribute.

        :Parameters:

            attr: `str`
                The attribute name.

        :Returns:

            `Query`
                The new query object.

        **Examples**

        >>> q = cf.eq(2001)
        >>> q
        <CF Query: (eq 2001)>
        >>> q = q.addattr('year')
        >>> q
        <CF Query: year(eq 2001)>

        >>> q = cf.lt(2)
        >>> q = q.addattr('A')
        >>> q = q.addattr('B')
        >>> q
        <CF Query: A.B(lt 2)>
        >>> q = q.addattr('C')
        >>> q
        <CF Query: A.B.C(lt 2)>

        """
        Q = type(self)
        new = Q.__new__(Q)

        new.__dict__ = self.__dict__.copy()
        new._attr += (attr,)

        new._NotImplemented_RHS_Data_op = True

        return new

    def copy(self):
        """Return a deep copy.

        ``q.copy()`` is equivalent to ``copy.deepcopy(q)``.

        :Returns:

                The deep copy.

        **Examples**

        >>> r = q.copy()

        """
        Q = type(self)
        new = Q.__new__(Q)

        d = self.__dict__.copy()
        new.__dict__ = d

        if d["_compound"]:
            d["_compound"] = deepcopy(d["_compound"])
        else:
            d["_value"] = deepcopy(d["_value"])

        return new

    @_display_or_return
    def dump(self, display=True):
        """Return a string containing a full description of the
        instance.

        :Parameters:

            display: `bool`, optional

                If `False` then return the description as a string. By
                default the description is printed.

        :Returns:

            `None` or `str`
                The description. If *display* is True then the description
                is printed and `None` is returned. Otherwise the
                description is returned as a string.

        """
        return str(self)

    @_deprecated_kwarg_check("traceback", version="3.0.0", removed_at="4.0.0")
    @_manage_log_level_via_verbosity
    def equals(self, other, verbose=None, traceback=False):
        """True if two `Query` objects are the same."""
        standard_difference_message = (
            f"{self.__class__.__name__}: Different compound components"
        )
        if self._compound:
            if not other._compound:
                logger.info(standard_difference_message)  # pragma: no cover
                return False

            if self._bitwise_operator != other._bitwise_operator:
                logger.info(
                    f"{self.__class__.__name__}: Different compound "
                    f"operators: {self._bitwise_operator!r}, "
                    f"{other._bitwise_operator!r}"
                )  # pragma: no cover
                return False

            if not self._compound[0].equals(other._compound[0]):
                if not self._compound[0].equals(other._compound[1]):
                    logger.info(
                        standard_difference_message
                    )  # pragma: no cover
                    return False
                if not self._compound[1].equals(other._compound[0]):
                    logger.info(
                        standard_difference_message
                    )  # pragma: no cover
                    return False
            elif not self._compound[1].equals(other._compound[1]):
                logger.info(standard_difference_message)  # pragma: no cover
                return False

        elif other._compound:
            logger.info(standard_difference_message)  # pragma: no cover
            return False

        for attr in (
            "_NotImplemented_RHS_Data_op",
            "_attr",
            "_value",
            "_operator",
        ):
            if not _equals(
                getattr(self, attr, None),
                getattr(other, attr, None),
                verbose=verbose,
            ):
                logger.info(
                    f"{self.__class__.__name__}: Different {attr!r} "
                    f"attributes: {getattr(self, attr, None)!r}, "
                    f"{getattr(other, attr, None)!r}"
                )  # pragma: no cover
                return False

        return True

    def evaluate(self, x):
        """Evaluate the query operation for a given left hand side
        operand.

        Note that for the query object ``q`` and any object, ``x``,
        ``x==q`` is equivalent to ``q.evaluate(x)`` and ``x!=q`` is
        equivalent to ``q.evaluate(x)==False``.

        :Parameters:

            x:
                The object for the left hand side operand of the query.

        :Returns:

                The result of the query. The nature of the result is
                dependent on the object type of *x*.

        **Examples**

        >>> q = cf.Query('lt', 5.5)
        >>> q.evaluate(6)
        False

        >>> q = cf.Query('wi', (1,2))
        >>> array = numpy.arange(4)
        >>> array
        array([0, 1, 2, 3])
        >>> q.evaluate(array)
        array([False,  True,  True, False], dtype=bool)

        """
        return self._evaluate(x, ())

    def _evaluate(self, x, parent_attr):
        """Evaluate the query operation for a given object.

        .. seealso:: `evaluate`

        :Parameters:

            x:
                See `evaluate`.

            parent_attr: `tuple`

        :Returns:

            See `evaluate`.

        """
        compound = self._compound
        attr = parent_attr + self._attr

        # ------------------------------------------------------------
        # Evaluate a compound condition
        # ------------------------------------------------------------
        if compound:
            c = compound[0]._evaluate(x, attr)
            d = compound[1]._evaluate(x, attr)
            return self._bitwise_operator(c, d)

        # ------------------------------------------------------------
        # Still here? Then evaluate a simple (non-compoundnd)
        # condition.
        # ------------------------------------------------------------
        for a in attr:
            x = getattr(x, a)

        operator = self._operator
        value = self._value

        standard_regex_error_msg = (
            f"Can't perform regular expression search on a non-string: {x!r}"
        )
        if operator == "eq":
            try:
                return bool(value.search(x))
            except AttributeError:
                return x == value
            except TypeError:
                raise ValueError(standard_regex_error_msg)

        if operator == "ne":
            try:
                return not bool(value.search(x))
            except AttributeError:
                return x != value
            except TypeError:
                raise ValueError(standard_regex_error_msg)

        if operator == "lt":
            _lt = getattr(x, "__query_lt__", None)
            if _lt is not None:
                return _lt(value)

            return x < value

        if operator == "le":
            _le = getattr(x, "__query_le__", None)
            if _le is not None:
                return _le(value)

            return x <= value

        if operator == "gt":
            _gt = getattr(x, "__query_gt__", None)
            if _gt is not None:
                return _gt(value)

            return x > value

        if operator == "ge":
            _ge = getattr(x, "__query_ge_", None)
            if _ge is not None:
                return _ge(value)

            return x >= value

        if operator == "wi":
            _wi = getattr(x, "__query_wi__", None)
            if _wi is not None:
                return _wi(value)

            return (x >= value[0]) & (x <= value[1])

        if operator == "wo":
            _wo = getattr(x, "__query_wo__", None)
            if _wo is not None:
                return _wo(value)

            return (x < value[0]) | (x > value[1])

        if operator == "set":
            if isinstance(x, str):
                for v in value:
                    try:
                        if v.search(x):
                            return True
                    except AttributeError:
                        if x == v:
                            return True

                return False
            else:
                _set = getattr(x, "__query_set__", None)
                if _set is not None:
                    return _set(value)

                i = iter(value)
                v = next(i)
                out = x == v
                for v in i:
                    out |= x == v

                return out

    def inspect(self):
        """Inspect the object for debugging.

        .. seealso:: `cf.inspect`

        :Returns:

            `None`

        """
        print(_inspect(self))  # pragma: no cover

    def iscontains(self):
        """Return True if the query is a "cell contains" condition.

        .. versionadded:: TODODASKVER

        .. seealso:: `cf.contains`

        :Returns:

            `bool`
                Whether or not the query is a "cell contains"
                condition.

        **Examples**

        >>> q = cf.contains(15)
        >>> q
        <CF Query: [lower_bounds(le 15) & upper_bounds(ge 15)]>
        >>> q.iscontains()
        True
        >>> q |= cf.lt(6)
        >>> q.iscontains()
        False

        >>> r = cf.wi(6, 10)
        >>> r.iscontains()
        False

        """
        if not (
            self._compound
            and self._value is not None
            and self._bitwise_operator is operator_and
        ):
            return False

        sig = builtin_set()
        for q in self._compound:
            if q._compound:
                # A compound query of compound queries is not a "cell
                # contains" condition
                break

            sig.add((q.attr, q.operator))

        return sig == builtin_set(
            ((("lower_bounds",), "le"), (("upper_bounds",), "ge"))
        )

    def set_condition_units(self, units):
        """Set units of condition values in-place.

        .. versionadded:: TODO

        :Parameters:

            units: `str` or `Units`

                The units to be set on all condition values.

        :Returns:

            `None`

        **Examples**

        >>> q = cf.lt(9)
        >>> q
        <CF Query: (lt 9)>
        >>> q.set_condition_units('km')
        >>> q
        <CF Query: (lt 9 km)>
        >>> q.set_condition_units('seconds')
            ...
        ValueError: Units <Units: seconds> are not equivalent to query condition units <Units: m>

        >>> q = cf.lt(9, units='m')
        >>> q
        <CF Query: (lt 9 m)>
        >>> q.set_condition_units('km')
        >>> q
        <CF Query: (lt 0.009 km)>

        >>> q = cf.lt(9)
        >>> r = cf.ge(3000, units='m')
        >>> s = q & r
        >>> s
        <CF Query: [(lt 9) & (ge 3000 m)]>
        >>> s.set_condition_units('km')
        >>> s
        <CF Query: [(lt 9 km) & (ge 3 km)]>
        >>> q
        <CF Query: (lt 9)>
        >>> r
        <CF Query: (ge 3000 m)>

        """

        def get_and_set_value_units(v, u):
            """Helper method to simplify setting of specified units."""
            v_units = getattr(v, "Units", None)
            if v_units is None:  # Value 'v' has no units
                v = Data(v, units=u)
            else:  # Value 'v' already has units
                try:
                    v = v.copy()
                    v.Units = u
                except ValueError:
                    raise ValueError(
                        f"Units {u!r} are not equivalent to "
                        f"query condition units {v_units!r}"
                    )

            return v

        units = Units(units)

        compound = self._compound
        if compound:
            for r in compound:
                r.set_condition_units(units)
            return

        value = self._value
        if value is None:
            return

        if self.operator in ("wi", "wo", "set"):
            # Value is a sequence of things that may or may not
            # already have units
            new_values = []
            for v in value:
                v = get_and_set_value_units(v, units)
                new_values.append(v)

            value = new_values
        else:
            value = get_and_set_value_units(value, units)

        self._value = value

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    @property
    def exact(self):
        """Deprecated at version 3.0.0.

        Use `re.compile` objects instead.

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "exact",
            "Use 're.compile' objects instead.",
            version="3.0.0",
            removed_at="4.0.0",
        )  # pragma: no cover

    def equivalent(self, other, traceback=False):
        """Deprecated at version 3.0.0."""
        _DEPRECATION_ERROR_FUNCTION(
            self, "equivalent", version="3.0.0", removed_at="4.0.0"
        )


# --------------------------------------------------------------------
# Constructor functions
# --------------------------------------------------------------------
def lt(value, units=None, attr=None):
    """A `Query` object for a "strictly less than" condition.

    .. seealso:: `cf.contains`, `cf.eq`, `cf.ge`, `cf.gt`, `cf.ne`,
                 `cf.le`, `cf.set`, `cf.wi`, `cf.wo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

        attr: `str`, optional
            Apply the condition to the attribute, or nested
            attributes, of the operand, rather than the operand
            itself. Nested attributes are specified by separating them
            with a ``.``. For example, the "month" attribute of the
            "bounds" attribute is specified as ``'bounds.month'``.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> c = cf.lt(5)
    >>> c == 2
    True
    >>> c == cf.Data(2, 'metres')
    <CF Data(): True>

    >>> c = cf.lt(5, 'metres')
    >>> c == 2
    True
    >>> c == cf.Data(50, 'centimetres')
    <CF Data(): True>

    >>> c = cf.lt(cf.Data(5, 'metres'))
    >>> c == 2
    True
    >>> c == cf.Data(50, 'centimetres')
    <CF Data(): True>

    >>> import numpy
    >>> c = cf.lt(numpy.array([2, 5]), attr='shape')
    >>> c
    <CF Query: shape(lt [2 5])>
    >>> c == numpy.arange(9).reshape(1, 9)
    array([ True, False])

    """
    return Query("lt", value, units=units, attr=attr)


def le(value, units=None, attr=None):
    """A `Query` object for a "less than or equal" condition.

    .. seealso:: `cf.contains`, `cf.eq`, `cf.ge`, `cf.gt`, `cf.ne`,
                 `cf.lt`, `cf.set`, `cf.wi`, `cf.wo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

        attr: `str`, optional
            Apply the condition to the attribute, or nested
            attributes, of the operand, rather than the operand
            itself. Nested attributes are specified by separating them
            with a ``.``. For example, the "month" attribute of the
            "bounds" attribute is specified as ``'bounds.month'``.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> q = cf.le(5)
    >>> q
    <CF Query: x le 5>
    >>> q.evaluate(5)
    True
    >>> q.evaluate(6)
    False

    """
    return Query("le", value, units=units, attr=attr)


def gt(value, units=None, attr=None):
    """A `Query` object for a "strictly greater than" condition.

    .. seealso:: `cf.contains`, `cf.eq`, `cf.ge`, `cf.ne`, `cf.le`,
                 `cf.lt`, `cf.set`, `cf.wi`, `cf.wo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

        attr: `str`, optional
            Apply the condition to the attribute, or nested
            attributes, of the operand, rather than the operand
            itself. Nested attributes are specified by separating them
            with a ``.``. For example, the "month" attribute of the
            "bounds" attribute is specified as ``'bounds.month'``.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> q = cf.gt(5)
    >>> q
    <CF Query: x gt 5>
    >>> q.evaluate(6)
    True
    >>> q.evaluate(5)
    False

    """
    return Query("gt", value, units=units, attr=attr)


def ge(value, units=None, attr=None):
    """A `Query` object for a "greater than or equal" condition.

    .. seealso:: `cf.contains`, `cf.eq`, `cf.gt`, `cf.ne`, `cf.le`,
                 `cf.lt`, `cf.set`, `cf.wi`, `cf.wo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

        attr: `str`, optional
            Apply the condition to the attribute, or nested
            attributes, of the operand, rather than the operand
            itself. Nested attributes are specified by separating them
            with a ``.``. For example, the "month" attribute of the
            "bounds" attribute is specified as ``'bounds.month'``.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> q = cf.ge(5)
    >>> q
    <CF Query: x ge 5>
    >>> q.evaluate(5)
    True
    >>> q.evaluate(4)
    False

    >>> cf.ge(10, 'm')
    <CF Query: (ge <CF Data: 10 m>)>
    >>> cf.ge(100, units=Units('kg'))
    <CF Query: (ge <CF Data: 100 kg>)>

    >>> cf.ge(2, attr='month')
    <CF Query: month(ge 2)>

    """
    return Query("ge", value, units=units, attr=attr)


def eq(value, units=None, attr=None, exact=True):
    """A `Query` object for an "equal" condition.

    .. seealso:: `cf.contains`, `cf.ge`, `cf.gt`, `cf.ne`, `cf.le`,
                 `cf.lt`, `cf.set`, `cf.wi`, `cf.wo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

        attr: `str`, optional
            Apply the condition to the attribute, or nested
            attributes, of the operand, rather than the operand
            itself. Nested attributes are specified by separating them
            with a ``.``. For example, the "month" attribute of the
            "bounds" attribute is specified as ``'bounds.month'``.

        exact: deprecated at version 3.0.0.
            Use `re.compile` objects in *value* instead.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> q = cf.eq(5)
    >>> q
    <CF Query: x eq 5>
    >>> q.evaluate(5)
    True
    >>> q == 4
    False

    >>> q = cf.eq('air')
    >>> q == 'air_temperature'
    True

    >>> q = cf.eq('.*temp')
    >>> q == 'air_temperature'
    True

    """
    if not exact:
        _DEPRECATION_ERROR_FUNCTION_KWARGS(
            "eq", exact=True
        )  # pragma: no cover

    return Query("eq", value, units=units, attr=attr)


def ne(value, units=None, attr=None, exact=True):
    """A `Query` object for a "not equal" condition.

    .. seealso:: `cf.contains`, `cf.eq`, `cf.ge`, `cf.gt`, `cf.le`,
                 `cf.lt`, `cf.set`, `cf.wi`, `cf.wo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

        attr: `str`, optional
            Apply the condition to the attribute, or nested
            attributes, of the operand, rather than the operand
            itself. Nested attributes are specified by separating them
            with a ``.``. For example, the "month" attribute of the
            "bounds" attribute is specified as ``'bounds.month'``.

        exact: deprecated at version 3.0.0.
            Use `re.compile` objects in *value* instead.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> q = cf.ne(5)
    >>> q
    <CF Query: x ne 5>
    >>> q.evaluate(4)
    True
    >>> q.evaluate(5)
    False

    """
    if not exact:
        _DEPRECATION_ERROR_FUNCTION_KWARGS(
            "ne", exact=True
        )  # pragma: no cover

    return Query("ne", value, units=units, attr=attr)


def wi(value0, value1, units=None, attr=None):
    """A `Query` object for a "within a range" condition.

    .. seealso:: `cf.contains`, `cf.eq`, `cf.ge`, `cf.gt`, `cf.ne`,
                 `cf.le`, `cf.lt`, `cf.set`, `cf.wo`

    :Parameters:

        value0:
             The lower bound of the range.

        value1:
             The upper bound of the range.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

        attr: `str`, optional
            Apply the condition to the attribute, or nested
            attributes, of the operand, rather than the operand
            itself. Nested attributes are specified by separating them
            with a ``.``. For example, the "month" attribute of the
            "bounds" attribute is specified as ``'bounds.month'``.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> q = cf.wi(5, 7)
    >>> q
    <CF Query: wi (5, 7)>
    >>> q.evaluate(6)
    True
    >>> q.evaluate(4)
    False

    """
    return Query("wi", [value0, value1], units=units, attr=attr)


def wo(value0, value1, units=None, attr=None):
    """A `Query` object for a "without a range" condition.

    .. seealso:: `cf.contains`, `cf.eq`, `cf.ge`, `cf.gt`, `cf.ne`,
                 `cf.le`, `cf.lt`, `cf.set`, `cf.wi`

    :Parameters:

        value0:
             The lower bound of the range.

        value1:
             The upper bound of the range.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

        attr: `str`, optional
            Apply the condition to the attribute, or nested
            attributes, of the operand, rather than the operand
            itself. Nested attributes are specified by separating them
            with a ``.``. For example, the "month" attribute of the
            "bounds" attribute is specified as ``'bounds.month'``.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> q = cf.wo(5)
    >>> q
    <CF Query: x wo (5, 7)>
    >>> q.evaluate(4)
    True
    >>> q.evaluate(6)
    False

    """
    return Query("wo", [value0, value1], units=units, attr=attr)


def set(values, units=None, attr=None, exact=True):
    """A `Query` object for a "member of set" condition.

    .. seealso:: `cf.contains`, `cf.eq`, `cf.ge`, `cf.gt`, `cf.ne`,
                 `cf.le`, `cf.lt`, `cf.wi`, `cf.wo`

    :Parameters:

        value: sequence
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

        attr: `str`, optional
            Apply the condition to the attribute, or nested
            attributes, of the operand, rather than the operand
            itself. Nested attributes are specified by separating them
            with a ``.``. For example, the "month" attribute of the
            "bounds" attribute is specified as ``'bounds.month'``.

        exact: deprecated at version 3.0.0.
            Use `re.compile` objects in *value* instead.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> c = cf.set([3, 5])
    >>> c == 4
    False
    >>> c == 5
    True
    >>> c == numpy.array([2, 3, 4, 5])
    array([False  True False  True])

    """
    if not exact:
        _DEPRECATION_ERROR_FUNCTION_KWARGS(
            "set", exact=True
        )  # pragma: no cover

    return Query("set", values, units=units, attr=attr)


def contains(value, units=None):
    """A `Query` object for a "cell contains" condition.

    .. versionadded:: 3.0.0

    .. seealso:: `cf.Query.iscontains`, `cf.cellsize`, `cf.cellge`,
                 `cf.cellgt`, `cf.cellne`, `cf.cellle`, `cf.celllt`,
                 `cf.cellwi`, `cf.cellwo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> q = cf.contains(8)
    >>> q
    <CF Query: [lower_bounds(le 8) & upper_bounds(ge 8)]>
    >>> q.value
    8

    >>> q = cf.contains(30, 'degrees_east')
    >>> q
    <CF Query: [lower_bounds(le 30 degrees_east) & upper_bounds(ge 30 degrees_east)]>
    >>> q.value
    <CF Data(): 30 degrees_east>

    >>> cf.contains(cf.Data(10, 'km'))
    <CF Query: [lower_bounds(le 10 km) & upper_bounds(ge 10 km)]>

    >>> c
    <CF DimensionCoordinate: longitude(4) degrees_east>
    >>> print(c.bounds.array)
    [[  0   90]
     [ 90  180]
     [180  270]
     [270  360]]
    >>> print((cf.contains(100) == c).array)
    [False True False False]
    >>> print((cf.contains(9999) == c).array)
    [False False False False]

    """
    return Query("le", value, units=units, attr="lower_bounds") & Query(
        "ge", value, units=units, attr="upper_bounds"
    )


def year(value):
    """A `Query` object for a "year" condition.

    .. seealso:: `cf.year`, `cf.month`, `cf.day`, `cf.hour`,
                 `cf.minute`, `cf.second`, `cf.seasons`, `cf.djf`,
                 `cf.mam`, `cf.jja`, `cf.son`

    :Parameter:

        value:
            The query condition's value.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> d = cf.dt(2002, 6, 16)
    >>> d == cf.year(2002)
    True
    >>> d == cf.year(cf.le(2003))
    True
    >>> d == cf.year(2001)
    False
    >>> d == cf.year(cf.wi(2003, 2006))
    False

    """
    if isinstance(value, Query):
        return value.addattr("year")
    else:
        return Query("eq", value, attr="year")


def month(value):
    """A `Query` object for a "month of the year" condition.

    .. seealso:: `cf.year`, `cf.day`, `cf.hour`, `cf.minute`,
                 `cf.second`, `cf.seasons`, `cf.djf`, `cf.mam`,
                 `cf.jja`, `cf.son`

    :Parameter:

        value:
            The query condition's value.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> d = cf.dt(2002, 6, 16)
    >>> d == cf.month(6)
    True
    >>> d == cf.month(cf.le(7))
    True
    >>> d == cf.month(7)
    False
    >>> d == cf.month(cf.wi(1, 6))
    True

    """
    if isinstance(value, Query):
        return value.addattr("month")
    else:
        return Query("eq", value, attr="month")


def day(value):
    """A `Query` object for a "day of the month" condition.

    .. seealso:: `cf.year`, `cf.month`, `cf.hour`, `cf.minute`,
                 `cf.second`, `cf.seasons`, `cf.djf`, `cf.mam`,
                 `cf.jja`, `cf.son`

    :Parameter:

        value:
            The query condition's value.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> d = cf.dt(2002, 6, 16)
    >>> d == cf.day(16)
    True
    >>> d == cf.day(cf.le(19))
    True
    >>> d == cf.day(7)
    False
    >>> d == cf.day(cf.wi(1, 21))
    True

    """
    if isinstance(value, Query):
        return value.addattr("day")
    else:
        return Query("eq", value, attr="day")


def hour(value):
    """A `Query` object for a "hour of the day" condition.

    .. seealso:: `cf.year`, `cf.month`, `cf.day`, `cf.minute`,
                 `cf.second`, `cf.seasons`, `cf.djf`, `cf.mam`, `cf.jja`,
                 `cf.son`

    :Parameter:

        value:
            The query condition's value.

    :Returns:

        `Query`
            The query object.

    In this context, any object which has a `!hour` attribute is
    considered to be a date-time variable.

    If *value* is a `Query` object then ``cf.hour(value)`` is
    equivalent to ``value.addattr('hour')``. Otherwise
    ``cf.hour(value)`` is equivalent to ``cf.eq(value, attr='hour')``.

    .. seealso:: `cf.year`, `cf.month`, `cf.day`, `cf.minute`,
                 `cf.second`

    :Parameters:

        value:
           Either the value that the hour is to be compared with, or a
           `Query` object for testing the hour.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> d = cf.dt(2002, 6, 16, 18)
    >>> d == cf.hour(18)
    True
    >>> d == cf.hour(cf.le(19))
    True
    >>> d == cf.hour(7)
    False
    >>> d == cf.hour(cf.wi(6, 23))
    True

    """
    if isinstance(value, Query):
        return value.addattr("hour")
    else:
        return Query("eq", value, attr="hour")


def minute(value):
    """A `Query` object for a "minute of the hour" condition.

    .. seealso:: `cf.year`, `cf.month`, `cf.day`, `cf.hour`,
                 `cf.second`, `cf.seasons`, `cf.djf`, `cf.mam`,
                 `cf.jja`, `cf.son`

    :Parameter:

        value:
            The query condition's value.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> d = cf.dt(2002, 6, 16, 18, 30, 0)
    >>> d == cf.minute(30)
    True
    >>> d == cf.minute(cf.le(45))
    True
    >>> d == cf.minute(7)
    False
    >>> d == cf.minute(cf.wi(15, 45))
    True

    """
    if isinstance(value, Query):
        return value.addattr("minute")
    else:
        return Query("eq", value, attr="minute")


def second(value):
    """A `Query` object for a "second of the minute" condition.

    .. seealso:: `cf.year`, `cf.month`, `cf.day`, `cf.hour`,
                 `cf.minute`, `cf.seasons`, `cf.djf`, `cf.mam`,
                 `cf.jja`, `cf.son`

    :Parameter:

        value:
            The query condition's value.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> d = cf.dt(2002, 6, 16, 18, 30, 0)
    >>> d == cf.second(0)
    True
    >>> d == cf.second(cf.le(30))
    True
    >>> d == cf.second(30)
    False
    >>> d == cf.second(cf.wi(0, 30))
    True

    """
    if isinstance(value, Query):
        return value.addattr("second")
    else:
        return Query("eq", value, attr="second")


def cellsize(value, units=None):
    """A `Query` object for a "cell size" condition.

    .. seealso:: `cf.contains`, `cf.cellge`, `cf.cellgt`, `cf.cellne`,
                 `cf.cellle`, `cf.celllt`, `cf.cellwi`, `cf.cellwo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> cf.cellsize(cf.lt(5, 'km'))
    <CF Query: cellsize(lt <CF Data: 5 km>)>
    >>> cf.cellsize(5)
    <CF Query: cellsize(eq 5)>
    >>> cf.cellsize(cf.Data(5, 'km'))
    <CF Query: cellsize(eq <CF Data: 5 km>)>
    >>> cf.cellsize(5, units='km')
    <CF Query: cellsize(eq <CF Data: 5 km>)>

    """
    if isinstance(value, Query):
        return value.addattr("cellsize")
    else:
        return Query("eq", value, units=units, attr="cellsize")


def cellwi(value0, value1, units=None):
    """A `Query` object for a "cell bounds lie within range" condition.

    .. seealso:: `cf.cellsize`, `cf.contains`, `cf.cellge`,
                 `cf.cellgt`, `cf.cellne`, `cf.cellle`, `cf.celllt`,
                 `cf.cellwo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> cf.cellwi(cf.Data(5, 'km'), cf.Data(10, 'km'))
    <CF Query: [lower_bounds(ge 5 km) & upper_bounds(le 10 km)]>
    >>> cf.cellwi(5, 10, units="km")
    <CF Query: [lower_bounds(ge 5 km) & upper_bounds(le 10 km)]>
    >>> cf.cellwi(cf.Data(5, 'km'), cf.Data(10000, 'm'))
    <CF Query: [lower_bounds(ge 5 km) & upper_bounds(le 10000 m)]>
    >>> cf.cellwi(0.2, 0.3)
    <CF Query: [lower_bounds(ge 0.2) & upper_bounds(le 0.3)]>

    """
    return Query("ge", value0, units=units, attr="lower_bounds") & Query(
        "le", value1, units=units, attr="upper_bounds"
    )


def cellwo(value0, value1, units=None):
    """A `Query` object for a "cell bounds lie without range" condition.

    .. seealso:: `cf.cellsize`, `cf.contains`, `cf.cellge`,
                 `cf.cellgt`, `cf.cellne`, `cf.cellle`, `cf.celllt`,
                 `cf.cellwi`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> cf.cellwo(cf.Data(5, 'km'), cf.Data(10, 'km'))
    <CF Query: [lower_bounds(lt 5 km) & upper_bounds(gt 10 km)]>
    >>> cf.cellwo(5, 10, units="km")
    <CF Query: [lower_bounds(lt 5 km) & upper_bounds(gt 10 km)]>
    >>> cf.cellwo(cf.Data(5, 'km'), cf.Data(10000, 'm'))
    <CF Query: [lower_bounds(lt 5 km) & upper_bounds(gt 10000 m)]>
    >>> cf.cellwo(0.2, 0.3)
    <CF Query: [lower_bounds(lt 0.2) & upper_bounds(gt 0.3)]>

    """
    return Query("lt", value0, units=units, attr="lower_bounds") & Query(
        "gt", value1, units=units, attr="upper_bounds"
    )


def cellgt(value, units=None):
    """A `Query` object for a "cell bounds strictly greater than"
    condition.

    .. seealso:: `cf.cellsize`, `cf.contains`, `cf.cellge`
                 `cf.cellne`, `cf.cellle`, `cf.celllt`, `cf.cellwi`,
                 `cf.cellwo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> cf.cellgt(cf.Data(300, 'K'))
    <CF Query: lower_bounds(gt 300 K)>
    >>> cf.cellgt(300, units='K')
    <CF Query: lower_bounds(gt 300 K)>
    >>> cf.cellgt(300)
    <CF Query: lower_bounds(gt 300)>

    """
    return Query("gt", value, units=units, attr="lower_bounds")


def cellge(value, units=None):
    """A `Query` object for a "cell bounds greater than or equal"
    condition.

    .. seealso:: `cf.cellsize`, `cf.contains`, `cf.cellgt`,
                 `cf.cellne`, `cf.cellle`, `cf.celllt`, `cf.cellwi`,
                 `cf.cellwo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.


    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> cf.cellge(cf.Data(300, 'K'))
    <CF Query: lower_bounds(ge 300 K)>
    >>> cf.cellge(cf.Data(300, 'K'))
    <CF Query: lower_bounds(ge 300 K)>
    >>> cf.cellge(300)
    <CF Query: lower_bounds(ge 300)>

    """
    return Query("ge", value, units=units, attr="lower_bounds")


def celllt(value, units=None):
    """A `Query` object for a “cell bounds strictly less than”
    condition.

    .. seealso:: `cf.cellsize`, `cf.contains`, `cf.cellge`,
                 `cf.cellgt`, `cf.cellne`, `cf.cellle`, `cf.cellwi`,
                 `cf.cellwo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> cf.celllt(cf.Data(300, 'K'))
    <CF Query: upper_bounds(lt 300 K)>
    >>> cf.celllt(300, units='K')
    <CF Query: upper_bounds(lt 300 K)>
    >>> cf.celllt(300)
    <CF Query: upper_bounds(lt 300)>

    """
    return Query("lt", value, units=units, attr="upper_bounds")


def cellle(value, units=None):
    """A `Query` object for a "cell bounds less than or equal"
    condition.

    .. seealso:: `cf.cellsize`, `cf.contains`, `cf.cellge`,
                 `cf.cellgt`, `cf.cellne`, `cf.celllt`, `cf.cellwi`,
                 `cf.cellwo`

    :Parameters:

        value:
            The query condition's value.

        units: `str` or `Units`, optional
            The units of *value*. By default, the same units as the
            operand being tested are assumed, if applicable. If
            *units* is specified and *value* already has units (such
            as those attached to a `Data` object), then the pair of
            units must be equivalent.

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> cf.cellle(cf.Data(300, 'K'))
    <CF Query: upper_bounds(le 300 K)>
    >>> cf.cellle(300, units='K')
    <CF Query: upper_bounds(le 300 K)>
    >>> cf.cellle(300)
    <CF Query: upper_bounds(le 300)>

    """
    return Query("le", value, units=units, attr="upper_bounds")


def jja():
    """A `Query` object for a "month of year in June, July or August"
    condition.

    .. versionadded:: 1.0

    .. seealso:: `cf.year`, `cf.month`, `cf.day`, `cf.hour`,
                 `cf.minute`, `cf.second`, `cf.seasons`, `cf.djf`,
                 `cf.mam`, `cf.son`

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> f
    <CF Field: air_temperature(time(365), latitude(64), longitude(128)) K>
    >>> f.subspace(time=cf.jja())
    <CF Field: air_temperature(time(92), latitude(64), longitude(128)) K>

    """
    return Query("wi", (6, 8), attr="month")


def son():
    """A `Query` object for a "month of year in September, October,
    November" condition.

    .. versionadded:: 1.0

    .. seealso:: `cf.year`, `cf.month`, `cf.day`, `cf.hour`,
                 `cf.minute`, `cf.second`, `cf.seasons`, `cf.djf`,
                 `cf.mam`, `cf.jja`

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> f
    <CF Field: air_temperature(time(365), latitude(64), longitude(128)) K>
    >>> f.subspace(time=cf.son())
    <CF Field: air_temperature(time(91), latitude(64), longitude(128)) K>

    """
    return Query("wi", (9, 11), attr="month")


def djf():
    """A `Query` object for a "month of year in December, January,
    February" condition.

    .. versionadded:: 1.0

    .. seealso:: `cf.year`, `cf.month`, `cf.day`, `cf.hour`,
                 `cf.minute`, `cf.second`, `cf.seasons`, `cf.mam`,
                 `cf.jja`, `cf.son`

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> f
    <CF Field: air_temperature(time(365), latitude(64), longitude(128)) K>
    >>> f.subspace(time=cf.djf())
    <CF Field: air_temperature(time(90), latitude(64), longitude(128)) K>

    """
    q = Query("ge", 12) | Query("le", 2)
    return q.addattr("month")


def mam():
    """A `Query` object for a "month of year in March, April, May"
    condition.

    .. versionadded:: 1.0

    .. seealso:: `cf.year`, `cf.month`, `cf.day`, `cf.hour`,
                 `cf.minute`, `cf.second`, `cf.seasons`, `cf.djf`,
                 `cf.jja`, `cf.son`

    :Returns:

        `Query`
            The query object.

    **Examples**

    >>> f
    <CF Field: air_temperature(time(365), latitude(64), longitude(128)) K>
    >>> f.subspace(time=cf.mam())
    <CF Field: air_temperature(time(92), latitude(64), longitude(128)) K>

    """
    return Query("wi", (3, 5), attr="month")


def seasons(n=4, start=12):
    """A customisable list of `Query` objects for "seasons in a year"
    conditions.

    Note that any date-time that lies within a particular season will
    satisfy that query.

    .. versionadded:: 1.0

    .. seealso:: `cf.year`, `cf.month`, `cf.day`, `cf.hour`, `cf.minute`,
                 `cf.second`, `cf.djf`, `cf.mam`, `cf.jja`, `cf.son`

    TODO

    .. seealso:: `cf.mam`, `cf.jja`, `cf.son`, `cf.djf`

    :Parameters:

        n: `int`, optional
            The number of seasons in the year. By default there are
            four seasons.

        start: `int`, optional
            The start month of the first season of the year. By
            default this is 12 (December).

    :Returns:

        `list` of `Query`
            The query objects.

    **Examples**

    >>> cf.seasons()
    [<CF Query: month[(ge 12) | (le 2)]>,
     <CF Query: month(wi (3, 5))>,
     <CF Query: month(wi (6, 8))>,
     <CF Query: month(wi (9, 11))>]

    >>> cf.seasons(4, 1)
    [<CF Query: month(wi (1, 3))>,
     <CF Query: month(wi (4, 6))>,
     <CF Query: month(wi (7, 9))>,
     <CF Query: month(wi (10, 12))>]

    >>> cf.seasons(3, 6)
    [<CF Query: month(wi (6, 9))>,
     <CF Query: month[(ge 10) | (le 1)]>,
     <CF Query: month(wi (2, 5))>]

    >>> cf.seasons(3)
    [<CF Query: month[(ge 12) | (le 3)]>,
     <CF Query: month(wi (4, 7))>,
     <CF Query: month(wi (8, 11))>]

    >>> cf.seasons(3, 6)
    [<CF Query: month(wi (6, 9))>,
     <CF Query: month[(ge 10) | (le 1)]>,
     <CF Query: month(wi (2, 5))>]

    >>> cf.seasons(12)
    [<CF Query: month(eq 12)>,
     <CF Query: month(eq 1)>,
     <CF Query: month(eq 2)>,
     <CF Query: month(eq 3)>,
     <CF Query: month(eq 4)>,
     <CF Query: month(eq 5)>,
     <CF Query: month(eq 6)>,
     <CF Query: month(eq 7)>,
     <CF Query: month(eq 8)>,
     <CF Query: month(eq 9)>,
     <CF Query: month(eq 10)>,
     <CF Query: month(eq 11)>]

    >>> cf.seasons(1, 4)
    [<CF Query: month[(ge 4) | (le 3)]>]

    """
    if 12 % n:
        raise ValueError("Number of seasons must divide into 12. Got %s" % n)

    if not 1 <= start <= 12 or int(start) != start:
        raise ValueError(
            "Start month must be integer between 1 and 12. Got %s" % start
        )

    out = []

    inc = int(12 / n)

    start = int(start)

    m0 = start
    for i in range(int(n)):
        m1 = ((m0 + inc) % 12) - 1
        if not m1:
            m1 = 12
        elif m1 == -1:
            m1 = 11

        if m0 < m1:
            q = Query("wi", (m0, m1))
        elif m0 > m1:
            q = Query("ge", m0) | Query("le", m1)
        else:
            q = Query("eq", m0)

        out.append(q.addattr("month"))

        m0 = m1 + 1
        if m0 > 12:
            m0 = 1

    return out


# --------------------------------------------------------------------
# Deprecated functions
# --------------------------------------------------------------------
def dtge(*args, **kwargs):
    """Return a `Query` object for a variable being not earlier than a
    date-time.

    Deprecated at version 3.0.0. Use 'cf.ge' with a datetime object
    value instead.

    """
    _DEPRECATION_ERROR_FUNCTION(
        "dtge",
        "Use 'cf.ge' with a datetime object value instead.",
        version="3.0.0",
        removed_at="4.0.0",
    )  # pragma: no cover


def dtgt(*args, **kwargs):
    """Deprecated at version 3.0.0.

    Use 'cf.gt' with a datetime object value instead.

    """
    _DEPRECATION_ERROR_FUNCTION(
        "dtgt",
        "Use 'cf.gt' with a datetime object value instead.",
        version="3.0.0",
        removed_at="4.0.0",
    )  # pragma: no cover


def dtle(*args, **kwargs):
    """Deprecated at version 3.0.0.

    Use 'cf.le' with a datetime object value instead.

    """
    _DEPRECATION_ERROR_FUNCTION(
        "dtle",
        "Use 'cf.le' with a datetime object value instead.",
        version="3.0.0",
        removed_at="4.0.0",
    )  # pragma: no cover


def dtlt(*args, **kwargs):
    """Deprecated at version 3.0.0.

    Use 'cf.lt' with a datetime object value instead.

    """
    _DEPRECATION_ERROR_FUNCTION(
        "dtlt",
        "Use 'cf.lt' with a datetime object value instead.",
        version="3.0.0",
        removed_at="4.0.0",
    )  # pragma: no cover


def dteq(*args, **kwargs):
    """Deprecated at version 3.0.0.

    Use 'cf.eq' with a datetime object value instead.

    """
    _DEPRECATION_ERROR_FUNCTION(
        "dteq",
        "Use 'cf.eq' with a datetime object value instead.",
        version="3.0.0",
        removed_at="4.0.0",
    )  # pragma: no cover


def dtne(*args, **kwargs):
    """Deprecated at version 3.0.0.

    Use 'cf.ne' with a datetime object value instead.

    """
    _DEPRECATION_ERROR_FUNCTION(
        "dtne",
        "Use 'cf.ne' with a datetime object value instead.",
        version="3.0.0",
        removed_at="4.0.0",
    )  # pragma: no cover


def contain(value, units=None, attr=None):
    """Return a `Query` object for coordinate cells containing a value.

    Deprecated at version 3.0.0. Use function 'cf.contains' instead.

    """
    _DEPRECATION_ERROR_FUNCTION(
        "cf.contain",
        "Use function 'cf.contains' instead",
        version="3.0.0",
        removed_at="4.0.0",
    )  # pragma: no cover



import cfdm

from . import Bounds, mixin


class AuxiliaryCoordinate(
    mixin.Coordinate, mixin.PropertiesDataBounds, cfdm.AuxiliaryCoordinate
):
    """An auxiliary coordinate construct of the CF data model.

    An auxiliary coordinate construct provides information which
    locate the cells of the domain and which depend on a subset of the
    domain axis constructs. Auxiliary coordinate constructs have to be
    used, instead of dimension coordinate constructs, when a single
    domain axis requires more then one set of coordinate values, when
    coordinate values are not numeric, strictly monotonic, or contain
    missing values, or when they vary along more than one domain axis
    construct simultaneously. CF-netCDF auxiliary coordinate variables
    and non-numeric scalar coordinate variables correspond to
    auxiliary coordinate constructs.

    The auxiliary coordinate construct consists of a data array of the
    coordinate values which spans a subset of the domain axis
    constructs, an optional array of cell bounds recording the extents
    of each cell (stored in a `Bounds` object), and properties to
    describe the coordinates. An array of cell bounds spans the same
    domain axes as its coordinate array, with the addition of an extra
    dimension whose size is that of the number of vertices of each
    cell. This extra dimension does not correspond to a domain axis
    construct since it does not relate to an independent axis of the
    domain. Note that, for climatological time axes, the bounds are
    interpreted in a special way indicated by the cell method
    constructs.

    **NetCDF interface**

    {{netCDF variable}}

    The netCDF variable group structure may be accessed with the
    `nc_set_variable`, `nc_get_variable`, `nc_variable_groups`,
    `nc_clear_variable_groups` and `nc_set_variable_groups` methods.

    """

    def __new__(cls, *args, **kwargs):
        """Store component classes."""
        instance = super().__new__(cls)
        instance._Bounds = Bounds
        return instance

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)
